# -*- coding: utf-8 -*-
"""
_SetState_

Oracle implementation of Locations.SetState
Created on Mon Jun 18 14:09:57 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Locations.SetState import SetState as MySQLSetState

class SetState(MySQLSetState):
    """
    Same as MySQL version
    """
    pass



"""
Definition of valid status values for a request and valid status transitions.

"""

from future.utils import viewitems

# make this list to ensure insertion order here
REQUEST_START_STATE = "new"
REQUEST_STATE_TRANSITION = {
    REQUEST_START_STATE: [REQUEST_START_STATE,
                          "assignment-approved",
                          "rejected"],

    "assignment-approved": ["assigned",  # manual transition
                            "rejected"],  # manual transition

    "assigned": ["staging",
                 "aborted"],  # manual transition
    "staging": ["staged",
                "aborted"],  # manual transition
    "staged": ["acquired",
               "aborted",  # manual transition
               "failed"],
    "acquired": ["running-open",
                 "aborted",  # manual transition
                 "failed"],
    "running-open": ["running-closed",
                     "force-complete",  # manual transition
                     "aborted"],  # manual transition

    "running-closed": ["force-complete",  # manual transition
                       "completed",
                       "aborted"],  # manual transition

    "force-complete": ["completed"],

    "failed": ["rejected",  # manual transition
               "assigned"],  # manual transition

    "completed": ["closed-out",
                  "rejected"],  # manual transition

    "closed-out": ["announced", "rejected"],  # manual transition

    "announced": ["rejected", "normal-archived"],

    "aborted": ["aborted-completed"],

    "aborted-completed": ["aborted-archived"],

    "rejected": ["rejected-archived"],

    # final status
    "normal-archived": ["rejected-archived"],

    "aborted-archived": [],

    "rejected-archived": []
}

# List of requests that a human user can change a request to
REQUEST_HUMAN_STATES = ["assignment-approved",
                        "assigned",
                        "force-complete",
                        "closed-out",
                        "announced",
                        "rejected",
                        "aborted"]

ACTIVE_STATUS = ["new",
                 "assignment-approved",
                 "assigned",
                 "staging",
                 "staged",
                 "acquired",
                 "running-open",
                 "running-closed",
                 "failed",
                 "force-complete",
                 "completed",
                 "closed-out",
                 "announced",
                 "aborted",
                 "aborted-completed",
                 "rejected"]

### WMSTATS_JOB_INFO + WMSTATS_NO_JOB_INFO is meant to be equal to ACTIVE_STATUS
WMSTATS_JOB_INFO = ["running-open",
                    "running-closed",
                    "force-complete",
                    "completed",
                    "closed-out"]

WMSTATS_NO_JOB_INFO = ["new",
                       "assignment-approved",
                       "assigned",
                       "staging",
                       "staged",
                       "acquired",
                       "failed",
                       "announced",
                       "aborted",
                       "aborted-completed",
                       "rejected"]

### Used for monitoring in T0-WMStats. See: Services/WMStats/WMStatsReader
T0_ACTIVE_STATUS = ["new",
                    "Closed",
                    "Merge",
                    "Harvesting",
                    "Processing Done",
                    "AlcaSkim",
                    "completed"]

# if the state is not defined here (new) allows all the property to get
# states in the key is the source states (need to define source states instead of destination states for GUI update)
ALLOWED_ACTIONS_FOR_STATUS = {
    "new": ["RequestPriority"],
    "assignment-approved": ["RequestPriority", "Team", "SiteWhitelist", "SiteBlacklist",
                            "AcquisitionEra", "ProcessingString", "ProcessingVersion",
                            "Dashboard", "MergedLFNBase", "TrustSitelists",
                            "UnmergedLFNBase", "MinMergeSize", "MaxMergeSize",
                            "MaxMergeEvents", "BlockCloseMaxWaitTime",
                            "BlockCloseMaxFiles", "BlockCloseMaxEvents", "BlockCloseMaxSize",
                            "SoftTimeout", "GracePeriod",
                            "TrustPUSitelists", "CustodialSites",
                            "NonCustodialSites", "Override",
                            "SubscriptionPriority"],
    "assigned": ["RequestPriority"],
    "staging": ["RequestPriority", "SiteWhitelist", "SiteBlacklist"],
    "staged": ["RequestPriority"],
    "acquired": ["RequestPriority", "SiteWhitelist", "SiteBlacklist"],
    "running-open": ["RequestPriority", "SiteWhitelist", "SiteBlacklist"],
    "running-closed": ["RequestPriority"],
    "failed": [],
    "force-complete": [],
    "completed": [],
    "closed-out": [],
    "announced": [],
    "aborted": [],
    "aborted-completed": [],
    "rejected": [],
    "normal-archived": [],
    "aborted-archived": [],
    "rejected-archived": [],
}

# NOTE: We need to Explicitly add RequestStatus to reqArgsDiff, since it is
#       missing from the ALLOWED_ACTIONS_FOR_STATUS mapping. The alternative
#       would be to add it as allowed action to every state.
#       The same applies to few more, such as all the keys needed for the
#       workqueue_stat_validation() calls, but we do this only during
#       request parameters validation.
ALLOWED_ACTIONS_ALL_STATUS = ["RequestStatus"]

# NOTE: We need to explicitly add all stat keys during validation
#       They are needed for the workqueue_stat_validation() calls
ALLOWED_STAT_KEYS = ['total_jobs', 'input_lumis', 'input_events', 'input_num_files']

# Workflow state transition automatically controlled by ReqMgr2
### NOTE: order of this list matters and it's used for status transition
AUTO_TRANSITION = {"staged": ["acquired", "running-open", "running-closed", "completed"],
                   "acquired": ["running-open", "running-closed", "completed"],
                   "running-open": ["running-closed", "completed"],
                   "running-closed": ["completed"]}

# Workflow state transition automatically controlled by ReqMgr2
# Specific to workflows either aborted or force-completed
CANCEL_AUTO_TRANSITION = {"aborted": "aborted-completed",
                          "force-complete": "completed"}

# list of destination states which doesn't allow any additional argument update
STATES_ALLOW_ONLY_STATE_TRANSITION = [key for key, val in viewitems(ALLOWED_ACTIONS_FOR_STATUS) if len(val) == 0]
# each item from STATUS_TRANSITION is a dictionary with 1 item, the key
# is name of the status
REQUEST_STATE_LIST = list(REQUEST_STATE_TRANSITION)

ACTIVE_STATUS_FILTER = {"RequestStatus": ['assignment-approved', 'assigned', 'staging', 'staged',
                                          'failed', 'acquired', 'running-open', 'running-closed',
                                          'force-complete', 'completed', 'closed-out']}


def check_allowed_transition(preState, postState):
    stateList = REQUEST_STATE_TRANSITION.get(preState, [])
    if postState in stateList:
        return True
    else:
        return False


def get_modifiable_properties(status=None):
    """
    returns mondifiable property list by status.
    if status is not defined return dictionarly of all the status and property list
    TODO: Currently gets the result from hardcoded list. change to get from configuration or db
    """
    if status:
        allowedKeys = ALLOWED_ACTIONS_FOR_STATUS.get(status, 'all_attributes')
        return allowedKeys
    else:
        return ALLOWED_ACTIONS_FOR_STATUS


def get_protected_properties():
    """
    returns properties never be modified once request is created
    """
    return ["RequestName", "_id"]



from __future__ import print_function, division

from builtins import str, object

from future.utils import viewitems

import logging
import os
import pickle
import signal
import time
import traceback

import htcondor


class OutputObj(object):
    """
    Class used when AuthenticatedSubprocess is created with pickleOut
    It stores the output message of the subprocess, any output object provided
    and extra information for debugging purposes, such as the environment
    """

    def __init__(self, outputMessage, outputObj):
        self.outputMessage = outputMessage
        self.outputObj = outputObj
        self.environmentStr = ""
        for key, val in viewitems(os.environ):
            self.environmentStr += "%s=%s\n" % (key, val)


class AuthenticatedSubprocess(object):
    """
    Context manager for execution of condor commands in a forked, especially useful for
    those commands where a different proxy credential is needed.
    """

    def __init__(self, proxy=None, pickleOut=False, outputObj=None, logger=logging):
        """
        Basic setup of the context manager
        :param proxy: optional path to the proxy file
        :param pickleOut: boolean flag which enables pickled output data
        :param outputObj: structure storing the actual output of the htcondor call
        :param logger: logger object
        """
        self.proxy = proxy
        self.pickleOut = pickleOut
        self.outputObj = outputObj
        self.timedout = False
        self.logger = logger

    def __enter__(self):
        self.r, self.w = os.pipe()
        self.rpipe = os.fdopen(self.r, 'r')
        self.wpipe = os.fdopen(self.w, 'w')
        self.pid = os.fork()
        if self.pid == 0 and self.proxy:
            # CRAB case
            htcondor.SecMan().invalidateAllSessions()
            htcondor.param['SEC_CLIENT_AUTHENTICATION_METHODS'] = 'FS,GSI'
            htcondor.param['DELEGATE_FULL_JOB_GSI_CREDENTIALS'] = 'true'
            htcondor.param['DELEGATE_JOB_GSI_CREDENTIALS_LIFETIME'] = '0'
            os.environ['X509_USER_PROXY'] = self.proxy
            self.rpipe.close()
        elif self.pid == 0:
            self.rpipe.close()
        else:
            self.wpipe.close()
        return self.pid, self.rpipe

    def __exit__(self, a, b, c):
        if self.pid == 0:
            if a is None and b is None and c is None:
                if self.pickleOut:
                    oo = OutputObj("OK", self.outputObj)
                    self.wpipe.write(pickle.dumps(oo))
                else:
                    self.wpipe.write("OK")
                self.wpipe.close()
                os._exit(0)
            else:
                tracebackString = str('\n'.join(traceback.format_tb(c)))
                msg = "Trapped exception in AuthenticatedSubprocess.Fork: %s %s %s \n%s" % \
                      (a, b, c, tracebackString)
                if self.pickleOut:
                    oo = OutputObj(msg, self.outputObj)
                    self.wpipe.write(pickle.dumps(oo))
                else:
                    self.wpipe.write(msg)
                self.wpipe.close()
                os._exit(1)
        else:
            timestart = time.time()
            self.timedout = True
            while (time.time() - timestart) < 3600:
                res = os.waitpid(self.pid, os.WNOHANG)
                if res != (0, 0):
                    self.timedout = False
                    break
                time.sleep(0.100)
            if self.timedout:
                self.logger.warning(
                    "Subprocess with PID %s (executed in AuthenticatedSubprocess) timed out. Killing it." % self.pid)
                os.kill(self.pid, signal.SIGTERM)
                # we should probably wait again and send SIGKILL if the kill does not work



#!/usr/bin/env python
"""
WorkQueue EndPolicyInterface

"""




from WMCore.WorkQueue.Policy.PolicyInterface import PolicyInterface
from WMCore.WorkQueue.DataStructs.WorkQueueElementResult import WorkQueueElementResult as WQEResult

class EndPolicyInterface(PolicyInterface):
    """Interface for end policies"""
    def __init__(self, **args):
        PolicyInterface.__init__(self, **args)
        self.results = []

    def __call__(self, elements, parents):
        """Apply policy to given elements"""
        for parent in parents:
            elements_for_parent = []
            for element in elements:
                if element['ParentQueueId'] == parent.id:
                    elements_for_parent.append(element)
            result = WQEResult(ParentQueueId = parent.id,
                               ParentQueueElement = parent,
                               Elements = elements_for_parent)
            self.results.append(result)

        self.applyPolicy() # do plugin logic
        return self.results

    def applyPolicy(self):
        """Extend in sub classes for custom behaviour"""
        forceStatus = None
        for result in self.results:
            result['ParentQueueElement'].updateWithResult(result)
            # check for a cancellation request
            if result['ParentQueueElement'].isCancelRequested():
                forceStatus = 'CancelRequested'
        if forceStatus:
            # cancel whole request
            [x.__setitem__('Status', forceStatus) for x in self.results]



#!/usr/bin/env python
"""
_WMInit

Init class that can be used by external projects
that only use part of the libraries
"""
from __future__ import print_function

import logging
import os
import os.path
import sys
import threading
import traceback
import warnings
import wmcoredb

from WMCore.Configuration import loadConfigurationFile
from WMCore.DAOFactory import DAOFactory
from WMCore.Database.DBFactory import DBFactory
from WMCore.Database.Transaction import Transaction
from WMCore.WMBase import getWMBASE
from WMCore.WMException import WMException


class WMInitException(WMException):
    """
    WMInitException

    You should never, ever see one of these.
    I'm not optimistic that this will be the case.
    """


def connectToDB():
    """
    _connectToDB_

    Connect to the database specified in the WMAgent config.
    """
    if "WMAGENT_CONFIG" not in os.environ:
        print("Please set WMAGENT_CONFIG to point at your WMAgent configuration.")
        sys.exit(1)

    if not os.path.exists(os.environ["WMAGENT_CONFIG"]):
        print("Can't find config: %s" % os.environ["WMAGENT_CONFIG"])
        sys.exit(1)

    wmAgentConfig = loadConfigurationFile(os.environ["WMAGENT_CONFIG"])

    if not hasattr(wmAgentConfig, "CoreDatabase"):
        print("Your config is missing the CoreDatabase section.")
        sys.exit(1)

    socketLoc = getattr(wmAgentConfig.CoreDatabase, "socket", None)
    connectUrl = getattr(wmAgentConfig.CoreDatabase, "connectUrl", None)
    (dialect, _) = connectUrl.split(":", 1)

    myWMInit = WMInit()
    myWMInit.setDatabaseConnection(dbConfig=connectUrl, dialect=dialect,
                                   socketLoc=socketLoc)

    return


class WMInit(object):
    def __init__(self):
        return

    def getWMBASE(self):
        """ for those that don't want to use the static version"""
        return getWMBASE()

    def setLogging(self, logFile=None, logName=None, logLevel=logging.INFO, logExists=True):
        """
        Sets logging parameters, depending on the settings,
        this method will create a logging file.
        """
        # use logName as name for file is no log file is given
        if not logExists:
            logging.basicConfig(level=logLevel, \
                                format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s', \
                                datefmt='%m-%d %H:%M', \
                                filename='%s.log' % logFile, \
                                filemode='w')
            logging.debug("Log file ready")

        myThread = threading.currentThread()
        if logName != None:
            myThread.logger = logging.getLogger(logName)
        else:
            myThread.logger = logging.getLogger()

    def setDatabaseConnection(self, dbConfig, dialect, socketLoc=None):
        """
        Sets the default connection parameters, without having to worry
        much on what attributes need to be set. This is esepcially
        advantagous for developers of third party projects that want
        to use only parts of the WMCore lib.

        The class differentiates between different formats used by external
        projects. External project formats that are supported can activated
        it by setting the flavor flag.
        """
        myThread = threading.currentThread()
        if getattr(myThread, "dialect", None) != None:
            # Database is already initialized, we'll create a new
            # transaction and move on.
            if hasattr(myThread, "transaction"):
                if myThread.transaction != None:
                    myThread.transaction.commit()

            myThread.transaction = Transaction(myThread.dbi)
            return

        options = {}
        if dialect.lower() in ['mysql', 'mariadb']:
            dialect = 'mariadb'  # Both MySQL and MariaDB use the mariadb directory
            if socketLoc != None:
                options['unix_socket'] = socketLoc
        elif dialect.lower() == 'oracle':
            dialect = 'oracle'  # Keep lowercase for consistency
        elif dialect.lower() == 'http':
            dialect = 'CouchDB'
        else:
            msg = "Unsupported dialect %s !" % dialect
            logging.error(msg)
            raise WMInitException(msg)

        myThread.dialect = dialect
        myThread.logger = logging
        myThread.dbFactory = DBFactory(logging, dbConfig, options)
        myThread.dbi = myThread.dbFactory.connect()

        # The transaction object will begin a transaction as soon as it is
        # initialized.  I'd rather have the user handle that, so we'll commit
        # it here.
        myThread.transaction = Transaction(myThread.dbi)
        myThread.transaction.commit()

        return

    def setSchema(self, modules=None, params=None):
        """
        Creates the schema in the database based on the modules
        input.

        This method needs to have been preceded by the
        setDatabaseConnection.

        @deprecated: Use setSchemaFromModules instead
        """
        warnings.warn("setSchema is deprecated. Use setSchemaFromModules instead.", DeprecationWarning)

        # create a map of old to new SQL module names
        moduleMap = {
            'WMCore.WMBS': 'wmbs',
            'WMCore.ResourceControl': 'resourcecontrol',
            'WMCore.BossAir': 'bossair',
            'WMCore.Agent.Database': 'agent',
            'WMComponent.DBS3Buffer': 'dbs3buffer',
            'T0.WMBS': 'tier0',
            'WMQuality.TestDB': 'testdb'
        }

        # convert old module names to new format
        if modules:
            modules = [moduleMap.get(module, module) for module in modules]

        self.setSchemaFromModules(modules)

    def setSchemaFromModules(self, sqlModules):
        """
        Initialize database schema for one or more SQL packages.
        It finds out which dialect is being used and then looks for the
        appropriate SQL files in the sql directory.

        :param sqlModules: List of SQL database modules to be initialized.
            Current supported values are (default is 'wmbs'):
            - 'wmbs'
            - 'resourcecontrol'
            - 'bossair'
            - 'agent'
            - 'dbs3buffer'
            - 'tier0'
        """
        myThread = threading.currentThread()
        if not hasattr(myThread, 'dbi'):
            raise WMInitException("Database connection not initialized. Call setDatabaseConnection first.")

        # Get the database dialect
        dialect = myThread.dialect.lower()
        if dialect not in ['mariadb', 'oracle']:
            raise WMInitException(f"Unsupported database dialect: {dialect}")

        # Get the base directory (WMCore root)
        if os.environ.get('WMCORE_ROOT'):
            baseDir = os.environ['WMCORE_ROOT']
            logging.info("SQL base directory based on WMCORE_ROOT: %s", baseDir)
        else:
            baseDir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
            logging.info("SQL base directory based on WMInit-relative path: %s", baseDir)

        # Define the SQL files needed for each module and their dependencies
        # The order matters - modules that depend on others should come later
        moduleSQLFiles = {
            'wmbs': {
                'files': ['create_wmbs_tables.sql', 'create_wmbs_indexes.sql', 'initial_wmbs_data.sql'],
                'dependencies': []
            },
            'resourcecontrol': {
                'files': ['create_resourcecontrol.sql'],
                'dependencies': ['wmbs']
            },
            'bossair': {
                'files': ['create_bossair.sql'],
                'dependencies': ['wmbs']
            },
            'agent': {
                'files': ['create_agent.sql'],
                'dependencies': []
            },
            'dbs3buffer': {
                'files': ['create_dbs3buffer.sql'],
                'dependencies': ['wmbs']
            },
            'tier0': {
                'files': ['create_tier0_tables.sql', 'create_tier0_indexes.sql', 'create_tier0_functions.sql', 'initial_tier0_data.sql'],
                'dependencies': ['wmbs', 'dbs3buffer']
            },
            'testdb': {
                'files': ['create_testdb.sql'],
                'dependencies': []
            },
        }

        # Validate all requested modules exist
        for module in sqlModules:
            if module not in moduleSQLFiles:
                raise WMInitException(f"Unknown module: {module}")

        # Sort modules based on dependencies
        sorted_modules = []
        remaining_modules = set(sqlModules)

        while remaining_modules:
            # Find modules with no remaining dependencies
            ready_modules = [
                mod for mod in remaining_modules
                if all(dep in sorted_modules for dep in moduleSQLFiles[mod]['dependencies'])
            ]

            if not ready_modules:
                # Circular dependency detected
                raise WMInitException("Circular dependency detected in module dependencies")

            sorted_modules.extend(ready_modules)
            remaining_modules -= set(ready_modules)

        # Execute SQL files in dependency order
        for module in sorted_modules:
            logging.info("Executing SQL files for: %s", module)
            for sql_file in moduleSQLFiles[module]['files']:
                dialect_sql_file = wmcoredb.get_sql_file(module_name=module, file_name=sql_file, backend=dialect)

                # now execute each SQL statement
                for stmt in self._getSQLStatements(dialect_sql_file, dialect):
                    try:
                        myThread.dbi.processData(stmt)
                    except Exception as ex:
                        msg = f"Error executing SQL file {dialect_sql_file}. "
                        msg += f"Statement: {stmt}"
                        msg += str(ex)
                        raise WMInitException(msg)

        return

    def _getSQLStatements(self, sqlFile, dialect):
        """
        Return the SQL statements from the file.
        For MariaDB, it accepts the whole SQL file content in a single statement.
        For Oracle, it splits the SQL file content into statements using the slash (/) terminator
        when it appears as the first character in a line.
        """
        if not os.path.exists(sqlFile):
            raise WMInitException(f"SQL file not found: {sqlFile}")

        with open(sqlFile, 'r', encoding='utf-8') as f:
            sql = f.read()

        if dialect == 'mariadb':
            return [sql]
        elif dialect == 'oracle':
            statements = []
            current_statement = []

            for line in sql.split('\n'):
                line = line.strip()
                # Skip empty lines and comments
                if not line or line.startswith('--'):
                    continue

                # If we find a slash terminator at the start of a line
                if line == '/':
                    # Join all lines collected so far into a statement
                    stmt = '\n'.join(current_statement)
                    if stmt.strip():
                        statements.append(stmt)
                    current_statement = []
                else:
                    current_statement.append(line)

            # Add any remaining statement
            if current_statement:
                stmt = '\n'.join(current_statement)
                if stmt.strip():
                    statements.append(stmt)

            return statements
        else:
            raise WMInitException(f"Unsupported database dialect: {dialect}")

    def clearDatabase(self, modules=None):
        """
        Database deletion. Global, ignore modules.
        """
        myThread = threading.currentThread()
        if hasattr(myThread, 'transaction') and getattr(myThread.transaction, 'transaction', None):
            # Then we have an open transaction
            # We should try and close it first
            try:
                myThread.transaction.commit()
            except:
                try:
                    myThread.transaction.rollback()
                except:
                    pass

        # Setup the DAO
        daoFactory = DAOFactory(package="WMCore.Database",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)
        destroyDAO = daoFactory(classname="Destroy")

        # Actually run a transaction and delete the DB
        try:
            destroyDAO.execute()
        except Exception as ex:
            msg = "Critical error while attempting to delete entire DB!\n"
            msg += str(ex)
            msg += str(traceback.format_exc())
            logging.error(msg)
            raise WMInitException(msg)

        return

    def checkDatabaseContents(self):
        """
        _checkDatabaseContents_

        Check and see if anything is in the database.
        This should be called by methods about to build the schema to make sure
        that the DB itself is empty.
        """

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.Database",
                                logger=myThread.logger,
                                dbinterface=myThread.dbi)

        testDAO = daoFactory(classname="ListUserContent")

        result = testDAO.execute()
        myThread.dbi.engine.dispose()

        return result



#!/usr/bin/env python
"""
_StageOut_

Basic Emulator for StageOut Step

"""
from __future__ import print_function

from WMCore.WMSpec.Steps.Emulator import Emulator

class StageOut(Emulator):
    """
    _StageOut_

    Emulate the execution of a StageOut Step

    """
    def pre(self):
        """
        _pre_

        Emulate the stage out pre step

        """
        return None

    def execute(self):
        """
        _execute_

        Emulate StageOut execution

        """
        self.step.section_("execution")
        self.step.execution.exitStatus = 0
        self.step.section_("emulation")
        self.step.emulation.emulatedBy = str(self.__class__.__name__)

        print("Emulating StageOut Step")


    def post(self):
        """
        _post_

        Emulate post stage out

        """
        return None



#!/usr/bin/env python
"""
_ThreadPool_

A class used for createing thread pools.
To use this you need to use the ThreadSlave class
"""






from builtins import str

import base64
import logging
import random
import threading
import time
from Utils.Utilities import encodeUnicodeToBytes
import pickle

from WMCore.ThreadPool.WorkQueue import ThreadPool as Queue
from WMCore.WMFactory import WMFactory

from Utils.PythonVersion import PY3

class ThreadPool(Queue):
    """
    _ThreadPool_

    A class used for creating persistent thread pools.
    To use this you need to use the ThreadSlave class
    """

    def __init__(self, slaveModule, component, \
        threadPoolID= 'default', nrOfSlaves = 0):
        """
        _init_

        Initializes pool, and resets lost threads (e.g. lost during crash).

        """

        #Queue.__init__(self, slaves)
        Queue.__init__(self, [])
        self.component = component
        self.callQueue = 0
        # we augment the threadpool id with component name in case
        # we use separate talbes for the thread queues to prevent
        # name clashes.
        compName = self.component.config.Agent.componentName
        self.threadPoolId = 'tp_threadpool_'+ \
            compName+'_'+threadPoolID
        # if set to false the peristent thread pool table is
        # created per threadpool id.
        self.oneQueue = True
        # the size of the buffer to minimize single inserts in large tables.
        self.bufferSize = 2
        # the maximum number of slaves (threads) we allow.
        self.nrOfSlaves = nrOfSlaves

        self.slaveName = slaveModule.split(".")[-1]

        self.slaveFactory = WMFactory("slaveFactory", \
            slaveModule[0:slaveModule.rfind(slaveModule.split(".")[-1])-1])
        myThread = threading.currentThread()
        myThread.transaction.begin()
        factory = WMFactory("threadPool", "WMCore.ThreadPool."+ \
            myThread.dialect)
        self.query = factory.loadObject("Queries")

        # check which tables we need to use.
        self.poolTable = 'tp_threadpool'
        self.poolTableBufferIn = 'tp_threadpool_buffer_in'
        self.poolTableBufferOut = 'tp_threadpool_buffer_out'
        if not self.oneQueue:
            self.poolTable = self.threadPoolId
            self.poolTableBufferIn = 'tp_threadpool_'+\
                compName+'_'+\
                threadPoolID+'_buffer_in'
            self.poolTableBufferOut = 'tp_threadpool_'+\
                compName+'_'+\
                threadPoolID+'_buffer_out'
            self.query.insertThreadPoolTables(self.poolTable)
            self.query.insertThreadPoolTables(self.poolTableBufferIn)
            self.query.insertThreadPoolTables(self.poolTableBufferOut)

        # restore any threads that might have been lost during a crash
        # de register thread in database so we do not need to restore it.
        msg = "THREADPOOL: Resetting lost threads to queue status if any"
        #logging.info(msg)
        args = {'componentName' : compName, \
            'thread_pool_id' : self.threadPoolId}
        self.query.updateWorkStatus(args, self.poolTable)
        self.query.updateWorkStatus(args, self.poolTableBufferIn)
        self.query.updateWorkStatus(args, self.poolTableBufferOut)
        # get the lenghts of the queue as this our starting point.
        self.callQueue = self.countMessages()
        # we do commit as initalization falls outside the while loop
        # of the component.
        myThread.transaction.commit()


    def prepareSlave(self, slave):
        """
        Prepares a slave to make sure it accesses
        the proper tables and is associated to the
        correct pool.
        """
        slave.args['thread_pool_table'] = self.poolTable
        slave.args['thread_pool_table_buffer_in'] = self.poolTableBufferIn
        slave.args['thread_pool_table_buffer_out'] = self.poolTableBufferOut
        slave.args['thread_pool_id'] = self.threadPoolId
        slave.args['thread_pool_buffer_size'] = self.bufferSize
        slave.args['componentName'] = self.component.config.Agent.componentName
        slave.component = self.component
        return slave

    def countMessages(self):
        """
        Counts how many messages for this pool are in the database.
        This is used when the pool is initialized to deal with entries
        that might be made before a crash and to establish an in
        memory queue length
        """

        msgs = 0
        args = {'componentName' : self.component.config.Agent.componentName, \
            'thread_pool_id' : self.threadPoolId}
        msgs += self.query.getQueueLength(args, self.poolTable)
        msgs += self.query.getQueueLength(args, self.poolTableBufferIn)
        msgs += self.query.getQueueLength(args, self.poolTableBufferOut)
        return msgs

    def enqueue( self, key, *parameters ):
        """
        _enqueue_

        Add a new work item to the queue.
        This may result in threads being spawned if there are threads
        available.

        """
        self.lock.acquire()
        base64_encoder = base64.encodebytes if PY3 else base64.encodestring
        args = {'event': str(key),
                'component' : self.component.config.Agent.componentName,
                'payload' : base64_encoder(encodeUnicodeToBytes(pickle.dumps(parameters))),
                'thread_pool_id' : self.threadPoolId}
        myThread = threading.currentThread()
        myThread.transaction.begin()
        self.query.insertWork(args, self.poolTableBufferIn)
        # we need to commit here otherwise the thread transaction might not
        # see it. check if this buffer needs to be flushed.
        myThread.transaction.commit()
        myThread.transaction.begin()
        bufferSize = self.query.getQueueLength(\
            {'componentName' : self.component.config.Agent.componentName, \
             'thread_pool_id' : self.threadPoolId}, self.poolTableBufferIn)
        if bufferSize > self.bufferSize:
            self.query.moveWorkFromBufferIn(self.poolTableBufferIn, \
                self.poolTable)
        #FIXME: we should call the msgService finsih method here before
        #this commit so we know the event/payload is transferred to a thread.
        myThread.transaction.commit()
        #logging.info("THREADPOOL: Enqueued item")

        # enqueue the work item
        self.callQueue += 1
        # check if there is a slave in the queue (then resue it)
        thread = None
        if len( self.slaveQueue ):
            # There is an available server: spawn a thread
            slave = self.slaveQueue[0]
            del self.slaveQueue[0]
            # Increment the count of active threads
            self.activeCount += 1
            thread = threading.Thread( target = self.slaveThread, \
                args=(slave,) )
            thread.start()
        # check if we can instantiate more slaves.
        else:
            if self.activeCount < self.nrOfSlaves:
                # we can still create slaves.
                slave  = \
                    self.slaveFactory.loadObject(self.slaveName)
                slave = self.prepareSlave(slave)
                self.activeCount += 1
                thread = threading.Thread( target = self.slaveThread, \
                    args=(slave,) )
                thread.start()

        self.lock.release()

        #if thread != None:
            #thread.join()



    def slaveThread( self, slaveServer ):
        """
        _slaveThread_

        This thread executes the method for each work item.

        """
        #~ print "slave thread starting up"
        self.lock.acquire()
        assert( self.activeCount > 0 )
        exceptCount = 0
        while self.callQueue > 0:
            # Dequeue work
            # handle exceptions to deal with deadlock issues.
            slaveServer.sane = True
            try:
                key, parameters = slaveServer.retrieveWork()
                self.callQueue -= 1
                exceptCount = 0
            except Exception as ex:
                logging.error("Problem with retrieving work : %s", str(ex))
                logging.error("Trying to salvage it")
                slaveServer.sane = False
                exceptCount += 1
                #TODO: Fix this exception; it's not a good exception
                #It just dumps the problem if it's the last thread in the queue
                if self.callQueue == 1 and exceptCount > 5:
                    self.callQueue -= 1
                    logging.error("If we got here, we screwed up badly enough I'm dumping it")
            self.lock.release()
            #~ print "making a call..."
            if slaveServer.sane:
                results = slaveServer( *parameters )
                slaveServer.removeWork()
                self.lock.acquire()
                # FIXME: do we want to keep this in?
                self.resultsQueue.append( (key, results) )
                # Notify the main thread that results are available,
                # if it was waiting
                self.lock.notify()
            else:
                # sleep for some random time.
                time.sleep(random.randint(1, 5))
                self.lock.acquire()


        # This thread is all done: put it back in the queue
        #~ print "slave thread exiting"
        self.activeCount -= 1
        assert( self.activeCount >= 0 )
        self.slaveQueue.append( slaveServer )
        self.lock.release()



#!/usr/bin/env python
"""
_MarkInjectedWorkflows_

Oracle implementation of Workflow.MarkInjectedWorkflows

"""
__all__ = []

from WMCore.WMBS.MySQL.Workflow.MarkInjectedWorkflows import MarkInjectedWorkflows as MySQLMarkInjectedWorkflows

class MarkInjectedWorkflows(MySQLMarkInjectedWorkflows):
    """
    Oracle Template

    """



#!/usr/bin/env python
"""
_SiblingProcessingBased_

Launch jobs to run over a file once all other subscriptions that process the file
have completed processing it.
"""

import threading
import logging

from WMCore.WMBS.File import File

from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.Services.UUIDLib import makeUUID
from WMCore.DAOFactory import DAOFactory

class SiblingProcessingBased(JobFactory):
    def algorithm(self, *args, **kwargs):
        """
        _algorithm_

        Run the discovery query and generate jobs if we find enough files.
        """
        # This doesn't use a proxy
        self.grabByProxy = False

        filesPerJob = int(kwargs.get("files_per_job", 10))

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package = "WMCore.WMBS",
                                logger = myThread.logger,
                                dbinterface = myThread.dbi)

        fileAvail = daoFactory(classname = "Subscriptions.SiblingSubscriptionsComplete")
        completeFiles = fileAvail.execute(self.subscription["id"],
                                          conn = myThread.transaction.conn,
                                          transaction = True)

        self.subscription["fileset"].load()
        if self.subscription["fileset"].open == True:
            filesetClosed = False
        else:
            fileFailed = daoFactory(classname = "Subscriptions.SiblingSubscriptionsFailed")
            fileFailed.execute(self.subscription["id"],
                               self.subscription["fileset"].id,
                               conn = myThread.transaction.conn,
                               transaction = True)
            filesetClosed = True

        fileSites = {}
        foundFiles = []
        for completeFile in completeFiles:
            if completeFile["lfn"] not in foundFiles:
                foundFiles.append(completeFile["lfn"])
            else:
                continue

            if completeFile["pnn"] not in fileSites:
                fileSites[completeFile["pnn"]] = []

            fileSites[completeFile["pnn"]].append(completeFile)

        for siteName in fileSites:
            if len(fileSites[siteName]) < filesPerJob and not filesetClosed:
                continue

            self.newGroup()
            while len(fileSites[siteName]) >= filesPerJob:
                self.newJob(name = makeUUID())
                for jobFile in fileSites[siteName][0:filesPerJob]:
                    newFile = File(id = jobFile["id"], lfn = jobFile["lfn"],
                                   events = jobFile["events"])
                    newFile["locations"] = set([jobFile["pnn"]])
                    self.currentJob.addFile(newFile)

                fileSites[siteName] = fileSites[siteName][filesPerJob:]

            if filesetClosed and len(fileSites[siteName]) > 0:
                self.newJob(name = makeUUID())
                for jobFile in fileSites[siteName]:
                    newFile = File(id = jobFile["id"], lfn = jobFile["lfn"],
                                   events = jobFile["events"])
                    newFile["locations"] = set([jobFile["pnn"]])
                    self.currentJob.addFile(newFile)

        return



#!/usr/bin/env python
"""
_UpdateBlocks_

Oracle implementation of DBS3Buffer.UpdateBlocks
"""

from WMComponent.DBS3Buffer.MySQL.UpdateBlocks import UpdateBlocks as MySQLUpdateBlocks

class UpdateBlocks(MySQLUpdateBlocks):
    """
    Identical to MySQL version.
    """



"""
_UpdateLocation_

Oracle implementation of Jobs.UpdateLocation

Created on Apr 18, 2013

@author: dballest
"""

from WMCore.WMBS.MySQL.Jobs.UpdateLocation import UpdateLocation as MySQLUpdateLocation

class UpdateLocation(MySQLUpdateLocation):
    pass



#!/usr/bin/env python
"""
_Delete_

Oracle implementation of Masks.Delete
"""

__all__ = []



from WMCore.WMBS.MySQL.Masks.Delete import Delete as DeleteMasksMySQL

class Delete(DeleteMasksMySQL):
    sql = DeleteMasksMySQL.sql



#!/usr/bin/env python


"""
Checks for finished jobs
Upon finding jobs cleans out their logs.
"""
from __future__ import print_function




import logging
import threading

from WMCore.Agent.Harness import Harness

from WMComponent.JobArchiver.JobArchiverPoller import JobArchiverPoller

#from WMCore.WorkQueue.WorkQueue import WorkQueue




class JobArchiver(Harness):
    """
    Checks for finished jobs
    Upon finding jobs cleans out their logs.
    """

    def __init__(self, config):
        # call the base class
        Harness.__init__(self, config)
        self.pollTime = 1

        self.config = config

        print("JobArchiver.__init__")

    def preInitialization(self):
        """
        Handles the setup of the worker thread.

        """
        logging.info("JobArchiver.preInitialization")

        # Add event loop to worker manager
        myThread = threading.currentThread()

        pollInterval = self.config.JobArchiver.JobArchiverPoller.pollInterval
        logging.info("Setting poll interval to %s seconds" % pollInterval)
        myThread.workerThreadManager.addWorker(JobArchiverPoller(self.config), pollInterval)

        return



#!/usr/bin/env python
"""
The actual jobArchiver algorithm
"""
from __future__ import division

import logging
import os
import os.path
import shutil
import tarfile
import threading

from Utils.IteratorTools import grouper
from Utils.Timers import timeFunction
from Utils.wmcoreDTools import resetWatchdogTimer, moduleName
from WMCore.DAOFactory import DAOFactory
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.Services.ReqMgrAux.ReqMgrAux import isDrainMode
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMException import WMException
from WMCore.WorkQueue.WorkQueueExceptions import WorkQueueNoMatchingElements
from WMCore.WorkQueue.WorkQueueUtils import queueFromConfig
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread


class JobArchiverPollerException(WMException):
    """
    _JobArchiverPollerException_

    The Exception handler for the job archiver.
    """


class JobArchiverPoller(BaseWorkerThread):
    """
    Polls for Error Conditions, handles them
    """

    def __init__(self, config):
        """
        Initialise class members
        """
        BaseWorkerThread.__init__(self)
        self.config = config
        self.changeState = ChangeState(self.config)

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        self.loadAction = self.daoFactory(classname="Jobs.LoadFromIDWithWorkflow")

        # Variables
        self.numberOfJobsToCluster = getattr(self.config.JobArchiver,
                                             "numberOfJobsToCluster", 1000)
        self.numberOfJobsToArchive = getattr(self.config.JobArchiver,
                                             "numberOfJobsToArchive", 10000)

        try:
            self.logDir = getattr(config.JobArchiver, 'logDir',
                                  os.path.join(config.JobArchiver.componentDir, 'logDir'))
            if not os.path.isdir(self.logDir):
                os.makedirs(self.logDir)
        except Exception as ex:
            msg = "Unhandled exception while setting up logDir!\n"
            msg += str(ex)
            logging.exception(msg)
            raise JobArchiverPollerException(msg)

        self.tier0Mode = hasattr(config, "Tier0Feeder")

        try:
            if not self.tier0Mode:
                self.workQueue = queueFromConfig(self.config)
        except Exception as ex:
            msg = "Could not load workQueue"
            msg += str(ex)
            logging.error(msg)
            # raise JobArchiverPollerException(msg)

        return

    def setup(self, parameters):
        """
        Load DB objects required for queries
        """
        return

    def terminate(self, params):
        """
        _terminate_

        This function terminates the job after a final pass
        """
        logging.debug("terminating. doing one more pass before we die")
        self.algorithm(params)
        return

    @timeFunction
    def algorithm(self, parameters=None):
        """
        Performs the archiveJobs method, looking for each type of failure
        And deal with it as desired.
        """
        try:
            self.archiveJobs()
            self.pollForClosable()
            self.markInjected()
        except WMException:
            myThread = threading.currentThread()
            if getattr(myThread, 'transaction', None) is not None \
                    and getattr(myThread.transaction, 'transaction', None) is not None:
                myThread.transaction.rollback()
            raise
        except Exception as ex:
            myThread = threading.currentThread()
            msg = "Caught exception in JobArchiver\n"
            msg += str(ex)
            msg += "\n\n"
            if getattr(myThread, 'transaction', None) is not None \
                    and getattr(myThread.transaction, 'transaction', None) is not None:
                myThread.transaction.rollback()
            raise JobArchiverPollerException(msg)

        # Reset its own watchdog timer at the end of the run cycle
        logging.info(f"Resetting {moduleName(self)} watchdog timer.")
        if resetWatchdogTimer(self):
            logging.warning(f"Failed to reset {moduleName(self)} watchdog timer. The component might be restarted soon.")

        return

    def archiveJobs(self):
        """
        _archiveJobs_

        archiveJobs will handle the master task of looking for finished jobs,
        and running the code that cleans them out.
        """
        doneList = self.findFinishedJobs()
        logging.info("Found %i finished jobs to archive", len(doneList))

        jobCounter = 0
        for slicedList in grouper(doneList, 10000):
            self.cleanWorkArea(slicedList)

            successList = []
            failList = []
            killList = []
            for job in slicedList:
                if job["outcome"] == "success":
                    successList.append(job)
                elif job["outcome"] == "killed":
                    killList.append(job)
                else:
                    failList.append(job)

            myThread = threading.currentThread()
            myThread.transaction.begin()
            self.changeState.propagate(successList, "cleanout", "success")
            self.changeState.propagate(failList, "cleanout", "exhausted")
            self.changeState.propagate(killList, "cleanout", "killed")
            myThread.transaction.commit()

            jobCounter += len(slicedList)
            logging.info("Successfully archived %d jobs out of %d.", jobCounter, len(doneList))

    def findFinishedJobs(self):
        """
        _findFinishedJobs_

        Will actually, surprisingly, find finished jobs (i.e., jobs either exhausted or successful)
        """
        jobList = []

        jobListAction = self.daoFactory(classname="Jobs.GetAllJobs")
        jobList1 = jobListAction.execute(state="success", limitRows=self.numberOfJobsToArchive)
        jobList2 = jobListAction.execute(state="exhausted", limitRows=self.numberOfJobsToArchive)
        jobList3 = jobListAction.execute(state="killed", limitRows=self.numberOfJobsToArchive)

        jobList.extend(jobList1)
        jobList.extend(jobList2)
        jobList.extend(jobList3)

        if len(jobList) == 0:
            # Then nothing is ready
            return []

        # Put together a list of job IDs
        binds = []
        for jobID in jobList:
            binds.append({"jobid": jobID})

        results = self.loadAction.execute(jobID=binds)

        if not isinstance(results, list):
            results = [results]

        doneList = []

        for entry in results:
            # One job per entry
            tmpJob = Job(id=entry['id'])
            tmpJob.update(entry)
            doneList.append(tmpJob)

        return doneList

    def cleanWorkArea(self, doneList):
        """
        _cleanWorkArea_

        Upon workQueue realizing that a subscriptions is done, everything
        regarding those jobs is cleaned up.
        """

        for job in doneList:
            # print "About to clean cache for job %i" % (job['id'])
            self.cleanJobCache(job)

        return

    def cleanJobCache(self, job):
        """
        _cleanJobCache_

        Clears out any files still sticking around in the jobCache,
        tars up the contents and sends them off
        """

        cacheDir = job['cache_dir']

        if not cacheDir or not os.path.isdir(cacheDir):
            msg = "Could not find jobCacheDir %s" % (cacheDir)
            logging.error(msg)
            return

        cacheDirList = os.listdir(cacheDir)

        if cacheDirList == []:
            os.rmdir(cacheDir)
            return

        # Now we need to set up a final destination
        try:
            # Label all directories by workflow
            # Workflow better have a first character
            workflow = job['workflow']
            firstCharacter = workflow[0]
            jobFolder = 'JobCluster_%i' \
                        % (int(job['id'] / self.numberOfJobsToCluster))
            logDir = os.path.join(self.logDir, firstCharacter,
                                  workflow, jobFolder)
            if not os.path.exists(logDir):
                os.makedirs(logDir)
        except Exception as ex:
            msg = "Exception while trying to make output logDir\n"
            msg += str("logDir: %s\n" % (logDir))
            msg += str(ex)
            logging.error(msg)
            raise JobArchiverPollerException(msg)

        # Otherwise we have something in there
        try:
            tarName = 'Job_%i.tar.bz2' % (job['id'])
            with tarfile.open(name=os.path.join(logDir, tarName), mode='w:bz2') as tarball:
                for fileName in cacheDirList:
                    fullFile = os.path.join(cacheDir, fileName)
                    try:
                        tarball.add(name=fullFile, arcname='Job_%i/%s' % (job['id'], fileName))
                    except IOError:
                        logging.error('Cannot read %s, skipping', fullFile)
        except Exception as ex:
            msg = "Exception while opening and adding to a tarfile\n"
            msg += "Tarfile: %s\n" % os.path.join(logDir, tarName)
            msg += str(ex)
            logging.error(msg)
            logging.debug("cacheDirList: %s", cacheDirList)
            raise JobArchiverPollerException(msg)

        try:
            shutil.rmtree('%s' % (cacheDir), ignore_errors=True)
        except Exception as ex:
            msg = "Error while removing the old cache dir.\n"
            msg += "CacheDir: %s\n" % cacheDir
            msg += str(ex)
            logging.error(msg)
            raise JobArchiverPollerException(msg)

        return

    def markInjected(self):
        """
        _markInjected_

        Mark any workflows that have been fully injected as injected
        """

        if self.tier0Mode:
            logging.debug("Component will not check workflows for injection status")
            return

        myThread = threading.currentThread()
        getAction = self.daoFactory(classname="Workflow.GetInjectedWorkflows")
        markAction = self.daoFactory(classname="Workflow.MarkInjectedWorkflows")
        result = getAction.execute()

        # Get the drain mode status
        drainMode = isDrainMode(self.config)

        # Check each result to see if it is injected:
        injected = []
        for name in result:
            try:
                if self.workQueue.getWMBSInjectionStatus(name, drainMode):
                    injected.append(name)
            except WorkQueueNoMatchingElements:
                # workflow not known - free to cleanup
                injected.append(name)
            except Exception as ex:
                logging.exception("Injection status checking failed, investigate: %s", str(ex))

        logging.info("Found %d workflows to mark as injected", len(injected))
        # Now, mark as injected those that returned True
        if len(injected) > 0:
            myThread.transaction.begin()
            markAction.execute(names=injected, injected=True)
            myThread.transaction.commit()
        return

    def pollForClosable(self):
        """
        _pollForClosable_

        Search WMBS for filesets that can be closed and mark them as closed.
        """
        myThread = threading.currentThread()
        myThread.transaction.begin()

        closableFilesetDAO = self.daoFactory(classname="Fileset.ListClosable")
        closableFilesets = closableFilesetDAO.execute()
        logging.info("Found %d filesets to be closed", len(closableFilesets))

        for closableFileset in closableFilesets:
            openFileset = Fileset(id=closableFileset)
            openFileset.load()

            logging.debug("Closing fileset %s", openFileset.name)
            openFileset.markOpen(False)

        myThread.transaction.commit()



"""LogDB Exceptions"""

class LogDBError(Exception):
    """Standard error baseclass"""
    def __init__(self, error):
        Exception.__init__(self, error)
        self.msg = LogDBError.__class__.__name__
        self.error = error

    def __str__(self):
        return "%s: %s" % (self.msg, self.error)



#!/usr/bin/env python
"""
_LsfPlugin_

Plugin to support use of LSF queues (at CERN)

Does not support rerunnable jobs, ie. which are
automatically requeued by LSF with a different job id
"""

import os
import re
import errno
import time
import datetime
import socket
import logging
import subprocess

from WMCore.WMInit import getWMBASE
from WMCore.BossAir.Plugins.BasePlugin import BasePlugin
from WMCore.FwkJobReport.Report import Report

class LsfPlugin(BasePlugin):
    """
    _LsfPlugin_

    """
    @staticmethod
    def stateMap():
        """
        For a given name, return a global state

        """
        stateDict = {'New':  'Pending',
                     'PEND': 'Pending',
                     'PSUSP': 'Pending',
                     'WAIT': 'Pending',
                     'RUN': 'Running',
                     'USUSP': 'Running',
                     'SSUSP': 'Running',
                     'DONE': 'Complete',
                     'EXIT': 'Error',
                     'UNKWN': 'Error',
                     'ZOMBI': 'Error',
                     'Timeout' : 'Error'}

        return stateDict


    def __init__(self, config):

        self.config = config

        BasePlugin.__init__(self, config)

        self.packageDir = None
        self.unpacker = os.path.join(getWMBASE(),
                                     'WMCore/WMRuntime/Unpacker.py')
        self.agent = config.Agent.agentName
        self.sandbox = None
        self.scriptFile = None
        self.queue = None
        self.resourceReq = None
        self.jobGroup = None
        self.basePrio = getattr(config.BossAir, 'LsfBasePrio', 50)
        self.batchOutput = None

        return


    def submit(self, jobs, info=None):
        """
        _submit_

        Submit jobs for one subscription

        """
        # If we're here, then we have submitter components
        self.scriptFile = self.config.JobSubmitter.submitScript
        self.queue = self.config.JobSubmitter.LsfPluginQueue
        self.resourceReq = getattr(self.config.JobSubmitter, 'LsfPluginResourceReq', None)
        self.jobGroup = self.config.JobSubmitter.LsfPluginJobGroup
        self.batchOutput = getattr(self.config.JobSubmitter, 'LsfPluginBatchOutput', None)

        successfulJobs = []
        failedJobs = []

        if len(jobs) == 0:
            # Then we have nothing to do
            return successfulJobs, failedJobs


        # Now assume that what we get is the following; a mostly
        # unordered list of jobs with random sandboxes.
        # We intend to sort them by sandbox.

        submitDict = {}
        for job in jobs:
            sandbox = job['sandbox']
            if sandbox not in submitDict:
                submitDict[sandbox] = []
            submitDict[sandbox].append(job)


        # Now submit the bastards
        for sandbox in submitDict:
            jobList = submitDict.get(sandbox, [])
            while len(jobList) > 0:
                jobsReady = jobList[:self.config.JobSubmitter.jobsPerWorker]
                jobList = jobList[self.config.JobSubmitter.jobsPerWorker:]

                for job in jobsReady:

                    if job == {}:
                        # Then I don't know how we got here either
                        logging.error("Was passed a nonexistant job.  Ignoring")
                        continue

                    submitScript = self.makeSubmit(job)

                    if not submitScript:
                        # Then we got nothing
                        logging.error("No submit script made!")
                        return {'NoResult': [0]}

                    submitScriptFile = os.path.join(job['cache_dir'], "submit.sh")
                    with open(submitScriptFile, 'w') as handle:
                        handle.writelines(submitScript)

                    # make reasonable job name
                    jobName = "WMAgentJob"
                    regExpParser = re.compile('.*/JobCreator/JobCache/([^/]+)/[^/]+/.*')
                    match = regExpParser.match(job['cache_dir'])
                    if match != None:
                        jobName = "%s-%s" % (match.group(1), job['id'])

                    # //
                    # // Submit LSF job
                    # //
                    command = 'bsub'
                    command += ' -q %s' % self.queue

                    if self.resourceReq != None:
                        command += ' -R "%s"' % self.resourceReq

                    command += ' -g %s' % self.jobGroup
                    command += ' -J %s' % jobName

                    lsfLogDir = self.batchOutput
                    if lsfLogDir != None:
                        now = datetime.datetime.today()
                        lsfLogDir += '/%s' % now.strftime("%Y%m%d%H")
                        try:
                            os.mkdir(lsfLogDir)
                            logging.debug("Created directory %s", lsfLogDir)
                        except OSError as err:
                            # suppress LSF log unless it's about an already exisiting directory
                            if err.errno != errno.EEXIST or not os.path.isdir(lsfLogDir):
                                logging.error("Can't create directory %s, turning off LSF log", lsfLogDir)
                                lsfLogDir = None

                    if lsfLogDir == None:
                        command += ' -oo /dev/null'
                    else:
                        command += ' -oo %s/%s.%%J.out' % (lsfLogDir, jobName)

                    if 'priority' in job:
                        try:
                            prio = int(job['priority'])
                            command += ' -sp %i' % (self.basePrio + prio)
                        except (ValueError, TypeError):
                            logging.debug("Priority for job %i not castable to an int\n", job['id'])
                            logging.debug("Not setting priority")
                            logging.debug("Priority: %s", job['priority'])
                        except Exception as ex:
                            logging.debug("Got unhandled exception while setting priority for job %i\n", job['id'])
                            logging.debug(str(ex))
                            logging.debug("Not setting priority")

                    command += ' < %s' % submitScriptFile

                    logging.info("Submitting LSF job: %s", command)

                    p = subprocess.Popen(command, shell=True,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.STDOUT)

                    stdout = p.communicate()[0]
                    returncode = p.returncode

                    if returncode == 0:
                        # check for correct naming convention in PFN
                        regExpParser = re.compile('Job <([0-9]+)> is submitted to queue')
                        match = regExpParser.search(stdout)
                        if match != None:
                            job['gridid'] = match.group(1)
                            successfulJobs.append(job)
                            logging.info("LSF Job ID : %s", job['gridid'])
                            continue
                        else:
                            logging.error("bsub didn't return a valid Job ID. Job is not submitted")
                            logging.error(stdout)

                    lsfErrorReport = Report()
                    lsfErrorReport.addError("JobSubmit", 61202, "LsfError", stdout)
                    job['fwjr'] = lsfErrorReport
                    failedJobs.append(job)

        # We must return a list of jobs successfully submitted,
        # and a list of jobs failed
        return successfulJobs, failedJobs


    def track(self, jobs):
        """
        _track_

        Track the jobs while in condor
        This returns a three-way ntuple
        First, the total number of jobs still running
        Second, the jobs that need to be changed
        Third, the jobs that need to be completed

        """
        # If we're here, then we have submitter components
        self.jobGroup = self.config.JobSubmitter.LsfPluginJobGroup

        changeList = []
        completeList = []
        runningList = []

        # get info about all active and recent jobs
        command = 'bjobs -a -w'
        command += ' -g %s' % self.jobGroup

        p = subprocess.Popen(command, shell=True,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        stdout = p.communicate()[0]
        returncode = p.returncode

        if returncode == 0:

            jobInfo = {}
            for line in stdout.splitlines(False)[1:]:

                # take line apart into elements
                linelist = line.rstrip().split()

                # dict with LSF jobid as key and LSF jobs status as value
                jobInfo[linelist[0]] = linelist[2]

            # now go over the jobs and see what we have
            for job in jobs:

                # if LSF doesn't know anything about the job, mark it complete
                if job['gridid'] not in jobInfo:
                    completeList.append(job)

                # otherwise act on LSF job status
                else:
                    newStatus = jobInfo[job['gridid']]

                    # track status changes
                    if newStatus != job['status']:
                        job['status'] = newStatus
                        job['status_time'] = int(time.time())
                        changeList.append(job)

                    job['globalState'] = LsfPlugin.stateMap()[newStatus]

                    # stop tracking finished jobs
                    if job['globalState'] in ['Complete', 'Error']:
                        completeList.append(job)
                    else:
                        runningList.append(job)

        return runningList, changeList, completeList


    def kill(self, jobs, raiseEx=False):
        """
        Kill a list of jobs based on their LSF jobid

        """

        for job in jobs:

            command = "bkill %s\n" % job['gridid']
            p = subprocess.Popen(command, shell=True,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.STDOUT)
            p.communicate()

        return


    def makeSubmit(self, job):
        """
        _makeSubmit_

        For a given job make a shell script to submit the job

        """

        script = ["#!/bin/sh\n"]

        # needed to construct rfio URL to access head node
        hostname = socket.getfqdn()

        # files needed to copied from head node to WN
        jobInputFiles = [job['sandbox'],
                         "%s/JobPackage.pkl" % job['packageDir'],
                         self.unpacker,
                         self.scriptFile]
        for filename in jobInputFiles:
            script.append("rfcp %s:%s .\n" % (hostname, filename))

        script.append("bash %s %s %s\n" % (os.path.basename(self.scriptFile),
                                           os.path.basename(job['sandbox']),
                                           job['id']))

        script.append("rfcp Report.%i.pkl %s:%s/\n" % (job["retry_count"], hostname, job['cache_dir']))

        # get back a lot of debug information to the head node
        #script.append("find . -type f -name '*.log' -exec rfcp {} %s:%s/ \;\n" % (hostname, job['cache_dir']))

        return script



#!/usr/bin/env python
"""
_ListJobStates_

Oracle implementation of Monitoring.ListJobStates
"""




from WMCore.WMBS.MySQL.Monitoring.ListJobStates import ListJobStates \
    as ListJobStatesMySQL

class ListJobStates(ListJobStatesMySQL):
    pass



"""
_UpdateWorkerError_

Oracle implementation of UpdateWorker
"""

__all__ = []



from WMCore.Agent.Database.MySQL.UpdateWorkerError import UpdateWorkerError \
     as UpdateWorkerErrorMySQL

class UpdateWorkerError(UpdateWorkerErrorMySQL):
    pass



"""
Perform cleanup actions
"""
import threading
import logging
import time

from Utils.Timers import timeFunction
from Utils.wmcoreDTools import resetWatchdogTimer, moduleName
from WMCore.WMFactory import WMFactory
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.Database.CMSCouch import CouchMonitor
from WMCore.Services.WorkQueue.WorkQueue import WorkQueue as WorkQueueService
from WMCore.Services.WMStats.WMStatsWriter import WMStatsWriter
from WMCore.Services.RequestDB.RequestDBWriter import RequestDBWriter
from WMComponent.AnalyticsDataCollector.DataCollectAPI import (LocalCouchDBData,
                                            WMAgentDBData, combineAnalyticsData,
                                            convertToRequestCouchDoc, initAgentInfo)

from WMComponent.DBS3Buffer.DBSBufferUtil import DBSBufferUtil


class AnalyticsPoller(BaseWorkerThread):
    """
    Gether the summary data for request (workflow) from local queue,
    local job couchdb, wmbs/boss air and populate summary db for monitoring
    """
    def __init__(self, config):
        """
        initialize properties specified from config
        """
        BaseWorkerThread.__init__(self)
        # set the workqueue service for REST call
        self.config = config
        # need to get campaign, user, owner info
        self.agentInfo = initAgentInfo(self.config)
        self.summaryLevel = (config.AnalyticsDataCollector.summaryLevel).lower()
        self.pluginName = getattr(config.AnalyticsDataCollector, "pluginName", None)
        self.plugin = None

    def setup(self, parameters):
        """
        set db connection(couchdb, wmbs) to prepare to gather information
        """
        # set the connection to local queue
        if not hasattr(self.config, "Tier0Feeder"):
            self.localQueue = WorkQueueService(self.config.AnalyticsDataCollector.localQueueURL)

        # set the connection for local couchDB call
        self.localCouchDB = LocalCouchDBData(self.config.AnalyticsDataCollector.localCouchURL,
                                             self.config.JobStateMachine.summaryStatsDBName,
                                             self.summaryLevel)

        # interface to WMBS/BossAir db
        myThread = threading.currentThread()
        # set wmagent db data
        self.wmagentDB = WMAgentDBData(self.summaryLevel, myThread.dbi, myThread.logger)
        # set the connection for local couchDB call
        self.localSummaryCouchDB = WMStatsWriter(self.config.AnalyticsDataCollector.localWMStatsURL,
                                                 appName="WMStatsAgent")

        # use local db for tier0
        if hasattr(self.config, "Tier0Feeder"):
            centralRequestCouchDBURL = self.config.AnalyticsDataCollector.localT0RequestDBURL
        else:
            centralRequestCouchDBURL = self.config.AnalyticsDataCollector.centralRequestDBURL

        self.centralRequestCouchDB = RequestDBWriter(centralRequestCouchDBURL,
                                                     couchapp=self.config.AnalyticsDataCollector.RequestCouchApp)
        self.centralWMStatsCouchDB = WMStatsWriter(self.config.General.centralWMStatsURL)

        #TODO: change the config to hold couch url
        self.localCouchServer = CouchMonitor(self.config.JobStateMachine.couchurl)

        self.dbsBufferUtil = DBSBufferUtil()

        if self.pluginName is not None:
            pluginFactory = WMFactory("plugins", "WMComponent.AnalyticsDataCollector.Plugins")
            self.plugin = pluginFactory.loadObject(classname=self.pluginName)

    @timeFunction
    def algorithm(self, parameters):
        """
        get information from wmbs, workqueue and local couch
        """
        try:
            # jobs per request info
            logging.info("Getting Job Couch Data ...")
            jobInfoFromCouch = self.localCouchDB.getJobSummaryByWorkflowAndSite()

            # fwjr per request info
            logging.info("Getting FWJRJob Couch Data ...")

            fwjrInfoFromCouch = self.localCouchDB.getJobPerformanceByTaskAndSiteFromSummaryDB()
            skippedInfoFromCouch = self.localCouchDB.getSkippedFilesSummaryByWorkflow()

            logging.info("Getting Batch Job Data ...")
            batchJobInfo = self.wmagentDB.getBatchJobInfo()

            logging.info("Getting Finished Task Data ...")
            finishedTasks = self.wmagentDB.getFinishedSubscriptionByTask()

            logging.info("Getting DBS PhEDEx upload status ...")
            completedWfs = self.dbsBufferUtil.getPhEDExDBSStatusForCompletedWorkflows(summary=True)

            # get the data from local workqueue:
            # request name, input dataset, inWMBS, inQueue
            logging.info("Getting Local Queue Data ...")
            localQInfo = {}
            if not hasattr(self.config, "Tier0Feeder"):
                localQInfo = self.localQueue.getAnalyticsData()
            else:
                logging.debug("Tier-0 instance, not checking WorkQueue")

            # combine all the data from 3 sources
            logging.info("""Combining data from
                                   Job Couch(%s),
                                   FWJR(%s),
                                   WorkflowsWithSkippedFile(%s),
                                   Batch Job(%s),
                                   Finished Tasks(%s),
                                   Local Queue(%s)
                                   Completed workflows(%s)..  ...""",
                         len(jobInfoFromCouch), len(fwjrInfoFromCouch), len(skippedInfoFromCouch),
                         len(batchJobInfo), len(finishedTasks), len(localQInfo), len(completedWfs))

            tempCombinedData = combineAnalyticsData(jobInfoFromCouch, batchJobInfo)
            tempCombinedData2 = combineAnalyticsData(tempCombinedData, localQInfo)
            combinedRequests = combineAnalyticsData(tempCombinedData2, completedWfs)

            # set the uploadTime - should be the same for all docs
            uploadTime = int(time.time())

            logging.info("%s requests Data combined,\n uploading request data...", len(combinedRequests))
            requestDocs = convertToRequestCouchDoc(combinedRequests, fwjrInfoFromCouch, finishedTasks,
                                                   skippedInfoFromCouch, self.agentInfo,
                                                   uploadTime, self.summaryLevel)

            if self.plugin != None:
                self.plugin(requestDocs, self.localSummaryCouchDB, self.centralRequestCouchDB)

            existingDocs = self.centralWMStatsCouchDB.getAllAgentRequestRevByID(self.agentInfo["agent_url"])
            self.centralWMStatsCouchDB.bulkUpdateData(requestDocs, existingDocs)

            logging.info("Request data upload success\n %s request, \nsleep for next cycle", len(requestDocs))

            self.centralWMStatsCouchDB.updateAgentInfoInPlace(self.agentInfo["agent_url"],
                                                              {"data_last_update": uploadTime, "data_error": "ok"})

        except Exception as ex:
            msg = str(ex)
            logging.exception("Error occurred, will retry later: %s", msg)

            try:
                self.centralWMStatsCouchDB.updateAgentInfoInPlace(self.agentInfo["agent_url"], {"data_error": msg})
            except:
                logging.error("upload Agent Info to central couch failed")

        # Reset its own watchdog timer at the end of the run cycle
        logging.info(f"Resetting {moduleName(self)} watchdog timer.")
        if resetWatchdogTimer(self):
            logging.warning(f"Failed to reset {moduleName(self)} watchdog timer. The component might be restarted soon.")



#!/usr/bin/env python

"""
MySQL implementation of GetChecksum
"""





from WMCore.Database.DBFormatter import DBFormatter

class GetChecksum(DBFormatter):
    sql = """SELECT cst.type AS cktype, fcs.cksum AS cksum FROM
               wmbs_file_checksums fcs INNER JOIN
               wmbs_checksum_type cst
               ON fcs.typeid = cst.id
               WHERE fcs.fileid = :fileid"""

    def formatResult(self, result):
        """
        I need the result in a reasonable list.
        This will return None if there is no cksum

        """
        formattedResult = {}

        dictVersion = DBFormatter.formatDict(self, result)
        if isinstance(dictVersion, list):
            if len(dictVersion) == 0:
                #Then it's empty
                return None
            else:
                #Otherwise there are several, and we have to record each one
                #I don't know how to do this yet.
                tmpDict = {}
                for entry in dictVersion:
                    tmpDict.update({entry.get('cktype', 'Default'): entry.get('cksum', None)})
                formattedResult['checksums']  = tmpDict
        else:
            formattedResult['checksums']  = {'Default': dictVersion.get('cksum', None)}
            if formattedResult == {'Default': None}:
                #Then the thing was empty anyway
                return None

        return formattedResult


    def execute(self, fileid = None, bulkList = None, conn = None, transaction = False):

        if bulkList:
            #Would need to accept a bulk list of form [{fileid: fileid}]
            binds = bulkList
        else:
            binds = {'fileid': fileid}

        result = self.dbi.processData(self.sql, binds,
                                      conn = conn, transaction = transaction)

        return self.formatResult(result)



"""
_CheckComponentStatus_

Oracle implementation of CheckComponentStatus
"""

__all__ = []



from WMCore.Agent.Database.MySQL.CheckComponentStatus import CheckComponentStatus \
     as CheckComponentStatusMySQL

class CheckComponentStatus(CheckComponentStatusMySQL):
    pass



#!/usr/bin/env python
"""
Script meant to fetch the number of blocks and files for a given dataset,
using both DBS and Rucio services. It prints any inconsistency among
those two.
"""


import logging
import sys
from collections import Counter

from WMCore.Services.DBS.DBS3Reader import DBS3Reader
from WMCore.Services.Rucio.Rucio import Rucio

RUCIO_ACCT = "wma_prod"
RUCIO_HOST = "http://cms-rucio.cern.ch"
RUCIO_AUTH = "https://cms-rucio-auth.cern.ch"
DBS_URL = "https://cmsweb-prod.cern.ch/dbs/prod/global/DBSReader"


def loggerSetup(logLevel=logging.INFO):
    """
    Return a logger which writes everything to stdout.
    """
    logger = logging.getLogger(__name__)
    outHandler = logging.StreamHandler(sys.stdout)
    outHandler.setFormatter(logging.Formatter("%(asctime)s:%(levelname)s:%(module)s: %(message)s"))
    outHandler.setLevel(logLevel)
    logger.addHandler(outHandler)
    logger.setLevel(logLevel)
    return logger


def getFromRucio(dataset, logger):
    """
    Using the WMCore Rucio object and fetch all the blocks and files
    for a given container.
    Returns a dictionary key'ed by the block name, value is the amount of files.
    """
    rucio = Rucio(acct=RUCIO_ACCT,
                  hostUrl=RUCIO_HOST,
                  authUrl=RUCIO_AUTH,
                  configDict={'logger': logger})

    result = dict()
    for block in rucio.getBlocksInContainer(dataset):
        data = rucio.getDID(block)
        result.setdefault(block, data['length'])
    return result


def getFromDBS(dataset, logger):
    """
    Uses the WMCore DBS3Reader object to fetch all the blocks and files
    for a given container.
    Returns a dictionary key'ed by the block name, and an inner dictionary
    with the number of valid and invalid files. It also returns a total counter
    for the number of valid and invalid files in the dataset.
    """
    dbsReader = DBS3Reader(DBS_URL, logger)

    result = dict()
    dbsFilesCounter = Counter({'valid': 0, 'invalid': 0})
    blocks = dbsReader.listFileBlocks(dataset)
    for block in blocks:
        data = dbsReader.dbs.listFileArray(block_name=block, validFileOnly=0, detail=True)
        result.setdefault(block, Counter({'valid': 0, 'invalid': 0}))
        for fileInfo in data:
            if fileInfo['is_file_valid'] == 1:
                result[block]['valid'] += 1
                dbsFilesCounter['valid'] += 1
            else:
                result[block]['invalid'] += 1
                dbsFilesCounter['invalid'] += 1
    return result, dbsFilesCounter


def main():
    """
    Expects a dataset name as input argument.
    It then queries Rucio and DBS and compare their blocks and
    number of files.
    """
    if len(sys.argv) != 2:
        print("A dataset name must be provided in the command line")
        sys.exit(1)
    datasetName = sys.argv[1]

    logger = loggerSetup(logging.INFO)

    rucioOutput = getFromRucio(datasetName, logger)
    dbsOutput, dbsFilesCounter = getFromDBS(datasetName, logger)

    logger.info("*** Dataset: %s", datasetName)
    logger.info("Rucio file count : %s", sum(rucioOutput.values()))
    logger.info("DBS file count   : %s", dbsFilesCounter['valid'] + dbsFilesCounter['invalid'])
    logger.info(" - valid files   : %s", dbsFilesCounter['valid'])
    logger.info(" - invalid files : %s", dbsFilesCounter['invalid'])
    logger.info("Blocks in Rucio but not in DBS: %s", set(rucioOutput.keys()) - set(dbsOutput.keys()))
    logger.info("Blocks in DBS but not in Rucio: %s", set(dbsOutput.keys()) - set(rucioOutput.keys()))

    for blockname in rucioOutput:
        if blockname not in dbsOutput:
            logger.error("This block does not exist in DBS: %s", blockname)
            continue
        if rucioOutput[blockname] != sum(dbsOutput[blockname].values()):
            logger.warning("Block with file mismatch: %s", blockname)
            logger.warning("\tRucio: %s\t\tDBS: %s", rucioOutput[blockname], sum(dbsOutput[blockname].values()))


if __name__ == "__main__":
    sys.exit(main())



"""
This module contains example of CMSSW metrics and corresponding helper
function to convert real metrics values to their data-types
"""

CMSSW_METRICS = {
    "SystemMemory": {
        "Active(file)": 4004848,
        "KernelStack": 13760,
        "VmallocUsed": 89264,
        "Hugepagesize": 2048,
        "Buffers": 81144,
        "DirectMap4k": 145248,
        "Inactive(file)": 3464500,
        "PageTables": 54704,
        "Shmem": 101300,
        "SReclaimable": 4005812,
        "Slab": 4393020,
        "SUnreclaim": 387208,
        "HardwareCorrupted": 0,
        "VmallocTotal": -1,
        "Committed_AS": 10468632,
        "MemFree": 8409300,
        "HugePages_Total": 0,
        "AnonHugePages": 14336,
        "Inactive": 5438044,
        "Inactive(anon)": 1973544,
        "Cached": 7490532,
        "MemAvailable": 19492112,
        "SwapFree": 28181320,
        "HugePages_Free": 0,
        "Bounce": 0,
        "CmaFree": 0,
        "WritebackTmp": 0,
        "Active(anon)": 7233764,
        "Mapped": 555512,
        "CommitLimit": 45402068,
        "Percpu": 4224,
        "AnonPages": 9069956,
        "Active": 11238612,
        "MemTotal": 29986740,
        "Mlocked": 0,
        "VmallocChunk": -224904,
        "NFS_Unstable": 0,
        "SwapTotal": 30408700,
        "DirectMap2M": 5408768,
        "Unevictable": 0,
        "CmaTotal": 0,
        "HugePages_Surp": 0,
        "HugePages_Rsvd": 0,
        "DirectMap1G": 25165824,
        "Dirty": 233852,
        "SwapCached": 87876,
        "Writeback": 0
        },
    "SystemCPU": {
        "CPUModels": "Intel Core Processor (Broadwell, IBRS)",
        "averageCoreSpeed": 2194.898,
        "totalCPUs": 16,
        "cpusetCount": 16
        },
    "ProcessingSummary": {
        "NumberEvents": 3,
        "NumberBeginRunCalls": 1,
        "NumberBeginLumiCalls": 1
        },
    "XrdSiteStatistics": {
        "readv-numOperations": [
            {
                "site": "roma1.infn.it",
                "value": 2
                },
            {
                "site": "T2_IT_BARI",
                "value": 5
                },
            {
                "site": "recas.ba.infn.it",
                "value": 15
                },
            {
                "site": "cern.ch",
                "value": 1
                }
            ],
        "read-totalMegabytes": [
            {
                "site": "roma1.infn.it",
                "value": 0.455857
                },
            {
                "site": "T2_IT_BARI",
                "value": 1.36202
                },
            {
                "site": "recas.ba.infn.it",
                "value": 1.58517
                },
            {
                "site": "cern.ch",
                "value": 0
                }
            ],
        "readv-totalMegabytes": [
            {
                "site": "roma1.infn.it",
                "value": 17.8692
                },
            {
                "site": "T2_IT_BARI",
                "value": 6.18529
                },
            {
                "site": "recas.ba.infn.it",
                "value": 71.6814
                },
            {
                "site": "cern.ch",
                "value": 0.251508
                }
            ],
        "read-totalMsecs": [
            {
                "site": "roma1.infn.it",
                "value": 1820.49
                },
            {
                "site": "T2_IT_BARI",
                "value": 4115.96
                },
            {
                "site": "recas.ba.infn.it",
                "value": 2925.22
                },
            {
                "site": "cern.ch",
                "value": 0
                }
            ],
        "read-numOperations": [
            {
                "site": "roma1.infn.it",
                "value": 72
                },
            {
                "site": "T2_IT_BARI",
                "value": 99
                },
            {
                "site": "recas.ba.infn.it",
                "value": 101
                },
            {
                "site": "cern.ch",
                "value": 0
                }
            ],
        "readv-totalMsecs": [
            {
                "site": "roma1.infn.it",
                "value": 5100.97
                },
            {
                "site": "T2_IT_BARI",
                "value": 1064.16
                },
            {
                "site": "recas.ba.infn.it",
                "value": 4575.52
                },
            {
                "site": "cern.ch",
                "value": 48.8356
                }
            ],
        "readv-numChunks": [
            {
                "site": "roma1.infn.it",
                "value": 108
                },
            {
                "site": "T2_IT_BARI",
                "value": 29
                },
            {
                "site": "recas.ba.infn.it",
                "value": 359
                },
            {
                "site": "cern.ch",
                "value": 13
                }
            ]
        },
    "Timing": {
        "MinEventTime": 10.3773,
        "NumberOfThreads": 2,
        "EventThroughput": 0.0379954,
        "NumberOfStreams": 2,
        "TotalEventSetupTime": 16.1899,
        "TotalLoopTime": 78.9568,
        "MaxEventTime": 46.6407,
        "TotalInitTime": 73.9733,
        "TotalJobChildrenCPU": 0.174912,
        "TotalJobTime": 153.001,
        "TotalNonModuleTime": 32.1743,
        "TotalInitCPU": 47.1554,
        "AvgEventTime": 34.1318,
        "TotalJobCPU": 155.483,
        "TotalLoopCPU": 108.262
        },
    "ApplicationMemory": {
        "SecondLargestRssEvent-f-RSS": 3319.17,
        "SecondLargestVsizeEventT2-f-RSS": 3284.8,
        "HEAP_MAPPED_SIZE_BYTES": 0,
        "LargestVsizeEventT1-c-EVENT": 102,
        "SecondLargestVsizeEventT2-a-COUNT": 2,
        "SecondLargestVsizeEventT2-e-DELTV": 0,
        "LargestVsizeIncreaseEvent-e-DELTV": 0,
        "HEAP_UNUSED_BYTES": 0,
        "HEAP_USED_BYTES": 0,
        "HEAP_MAPPED_N_CHUNKS": 0,
        "PeakValueRss": 3331.98,
        "ThirdLargestRssEvent-a-COUNT": 2,
        "LargestVsizeEventT1-f-RSS": 3331.98,
        "ThirdLargestVsizeEventT3-c-EVENT": 101,
        "ThirdLargestVsizeEventT3-e-DELTV": 0,
        "SecondLargestRssEvent-e-DELTV": 0,
        "LargestVsizeEventT1-a-COUNT": 1,
        "ThirdLargestRssEvent-b-RUN": 1,
        "HEAP_ARENA_N_UNUSED_CHUNKS": 1,
        "SecondLargestRssEvent-a-COUNT": 3,
        "AverageGrowthRateRss": 669.988,
        "ThirdLargestVsizeEventT3-d-VSIZE": 5453.68,
        "AverageGrowthRateVsize": 895.158,
        "LargestVsizeEventT1-e-DELTV": 0,
        "LargestVsizeEventT1-d-VSIZE": 5453.68,
        "LargestRssEvent-d-VSIZE": 5453.68,
        "PeakValueVsize": 5453.68,
        "LargestRssEvent-a-COUNT": 1,
        "SecondLargestVsizeEventT2-d-VSIZE": 5453.68,
        "LargestVsizeIncreaseEvent-f-RSS": 3331.98,
        "ThirdLargestRssEvent-c-EVENT": 104,
        "LargestVsizeEventT1-b-RUN": 1,
        "LargestVsizeIncreaseEvent-a-COUNT": 1,
        "LargestRssEvent-b-RUN": 1,
        "ThirdLargestRssEvent-f-RSS": 3284.8,
        "ThirdLargestVsizeEventT3-a-COUNT": 3,
        "LargestVsizeIncreaseEvent-d-VSIZE": 5453.68,
        "SecondLargestRssEvent-b-RUN": 1,
        "ThirdLargestVsizeEventT3-f-RSS": 3319.17,
        "ThirdLargestRssEvent-d-VSIZE": 5453.68,
        "LargestRssEvent-c-EVENT": 102,
        "HEAP_TOP_FREE_BYTES": 0,
        "HEAP_ARENA_SIZE_BYTES": 0,
        "LargestRssEvent-f-RSS": 3331.98,
        "SecondLargestVsizeEventT2-c-EVENT": 104,
        "LargestRssEvent-e-DELTV": 0,
        "LargestVsizeIncreaseEvent-b-RUN": 1,
        "SecondLargestVsizeEventT2-b-RUN": 1,
        "ThirdLargestVsizeEventT3-b-RUN": 1,
        "SecondLargestRssEvent-d-VSIZE": 5453.68,
        "ThirdLargestRssEvent-e-DELTV": 0,
        "LargestVsizeIncreaseEvent-c-EVENT": 102,
        "SecondLargestRssEvent-c-EVENT": 101
        },
    "StorageStatistics": {
        "Timing-tstoragefile-stat-maxMsecs": 0.072352,
        "Parameter-untracked-bool-enabled": True,
        "Timing-tstoragefile-flush-minMsecs": 0.646707,
        "Timing-root-close-minMsecs": 0.10421,
        "Timing-tstoragefile-flush-totalMegabytes": 0,
        "Timing-file-readv-maxMsecs": 0,
        "Timing-file-close-numSuccessfulOperations": 1,
        "Timing-file-read-numOperations": 4680,
        "Timing-file-read-totalMsecs": 61.8403,
        "Timing-tstoragefile-writeViaCache-numOperations": 7307,
        "Timing-file-flush-numOperations": 3,
        "Timing-file-prefetch-numOperations": 0,
        "Timing-file-close-totalMsecs": 0.190476,
        "Timing-tstoragefile-readAsync-minMsecs": 0,
        "Timing-root-open-numOperations": 3,
        "Timing-root-position-totalMegabytes": 0,
        "Timing-tstoragefile-readViaCache-minMsecs": 0,
        "Timing-file-construct-totalMsecs": 0.00055,
        "Timing-root-stagein-totalMsecs": 4.10695,
        "Timing-tstoragefile-flush-numSuccessfulOperations": 3,
        "Timing-tstoragefile-write-minMsecs": 0.006056,
        "Timing-tstoragefile-readAsync-totalMsecs": 0,
        "Timing-file-write-totalMegabytes": 29.0666,
        "Timing-file-open-minMsecs": 1.05184,
        "Timing-root-open-numSuccessfulOperations": 3,
        "Timing-root-construct-numSuccessfulOperations": 3,
        "Timing-file-read-minMsecs": 0.001137,
        "Timing-file-close-maxMsecs": 0.190476,
        "Timing-tstoragefile-readActual-numSuccessfulOperations": 4968,
        "Timing-file-write-numSuccessfulOperations": 7307,
        "Timing-root-prefetch-minMsecs": 0,
        "Timing-root-stagein-numSuccessfulOperations": 1,
        "Timing-root-readv-numOperations": 231,
        "Timing-root-close-totalMsecs": 0.741968,
        "Timing-tstoragefile-construct-numSuccessfulOperations": 5,
        "Timing-tstoragefile-stat-totalMsecs": 2.83543,
        "Timing-file-flush-maxMsecs": 1809.22,
        "Timing-file-open-totalMegabytes": 0,
        "Timing-tstoragefile-readAsync-numOperations": 30,
        "Timing-root-position-numSuccessfulOperations": 93,
        "Timing-tstoragefile-write-numOperations": 7307,
        "Timing-root-read-numOperations": 57,
        "Timing-file-writev-totalMsecs": 0,
        "Timing-tstoragefile-readAsync-totalMegabytes": 0,
        "Timing-file-writev-maxMsecs": 0,
        "Timing-root-writev-totalMegabytes": 0,
        "Timing-tstoragefile-writeActual-totalMegabytes": 29.0666,
        "Timing-file-position-totalMsecs": 17.9375,
        "Timing-tstoragefile-readViaCache-maxMsecs": 0,
        "Timing-file-open-totalMsecs": 3.28384,
        "Timing-root-writev-numSuccessfulOperations": 0,
        "Timing-tstoragefile-writeViaCache-totalMsecs": 0,
        "Timing-file-prefetch-numSuccessfulOperations": 0,
        "Timing-tstoragefile-readViaCache-totalMegabytes": 0,
        "Timing-file-prefetch-totalMsecs": 0,
        "Timing-file-flush-numSuccessfulOperations": 3,
        "Timing-tstoragefile-stat-minMsecs": 0.00071,
        "Timing-file-position-totalMegabytes": 0,
        "Timing-root-prefetch-totalMegabytes": 0,
        "Timing-root-position-numOperations": 93,
        "Timing-tstoragefile-close-numOperations": 4,
        "Timing-file-position-minMsecs": 0.000751,
        "Timing-root-read-totalMsecs": 2011.12,
        "Timing-tstoragefile-construct-totalMegabytes": 0,
        "Parameter-untracked-bool-prefetching": False,
        "Timing-file-open-numOperations": 2,
        "Timing-tstoragefile-writeViaCache-numSuccessfulOperations": 0,
        "Timing-tstoragefile-writeActual-numSuccessfulOperations": 7307,
        "Timing-root-readv-minMsecs": 15.7082,
        "Timing-file-construct-totalMegabytes": 0,
        "Timing-root-position-minMsecs": "4.4e-05",
        "Timing-root-readv-maxMsecs": 4168.77,
        "Timing-tstoragefile-seek-totalMegabytes": 0,
        "Timing-root-close-numOperations": 3,
        "Timing-root-close-numSuccessfulOperations": 3,
        "Timing-root-readv-numSuccessfulOperations": 231,
        "Timing-file-flush-minMsecs": 0.645059,
        "Timing-tstoragefile-seek-minMsecs": 0.000217,
        "Timing-root-read-maxMsecs": 258.502,
        "Timing-file-write-maxMsecs": 12.2384,
        "Timing-root-write-numSuccessfulOperations": 0,
        "Timing-tstoragefile-readAsync-numSuccessfulOperations": 0,
        "Timing-root-read-minMsecs": 15.753,
        "Timing-file-readv-numSuccessfulOperations": 0,
        "Timing-file-readv-minMsecs": 0,
        "Timing-file-prefetch-maxMsecs": 0,
        "Timing-root-construct-totalMegabytes": 0,
        "Timing-tstoragefile-construct-numOperations": 5,
        "Timing-file-read-numSuccessfulOperations": 4680,
        "Timing-root-write-totalMsecs": 0,
        "Timing-file-readv-numOperations": 0,
        "Timing-root-prefetch-numSuccessfulOperations": 0,
        "Timing-file-prefetch-totalMegabytes": 0,
        "Timing-file-readv-totalMegabytes": 0,
        "Timing-root-readv-totalMegabytes": 96.7375,
        "Timing-tstoragefile-close-totalMegabytes": 0,
        "Timing-file-writev-minMsecs": 0,
        "Timing-file-construct-minMsecs": 0.000259,
        "Timing-tstoragefile-flush-numOperations": 3,
        "Timing-tstoragefile-readActual-minMsecs": 0.001295,
        "Timing-file-write-minMsecs": 0.005698,
        "ROOT-tfile-write-totalMegabytes": 0,
        "Timing-root-open-maxMsecs": 4379.9,
        "Timing-tstoragefile-readActual-numOperations": 4968,
        "Timing-tstoragefile-flush-maxMsecs": 1809.23,
        "Timing-root-close-maxMsecs": 0.402898,
        "Timing-root-open-totalMegabytes": 0,
        "Timing-tstoragefile-read-totalMegabytes": 8.08306,
        "Timing-root-position-totalMsecs": 0.032065,
        "Timing-file-open-numSuccessfulOperations": 2,
        "Timing-root-write-numOperations": 0,
        "Timing-tstoragefile-read-numOperations": 4737,
        "Timing-root-open-totalMsecs": 9945.23,
        "ROOT-tfile-read-totalMegabytes": 5.55608,
        "Parameter-untracked-string-cacheHint": "auto-detect",
        "Timing-root-write-maxMsecs": 0,
        "Timing-file-writev-numOperations": 0,
        "Timing-root-writev-numOperations": 0,
        "Timing-file-writev-numSuccessfulOperations": 0,
        "Timing-file-close-totalMegabytes": 0,
        "Timing-tstoragefile-write-totalMegabytes": 29.0666,
        "Timing-tstoragefile-close-maxMsecs": 0.40434,
        "Timing-file-flush-totalMegabytes": 0,
        "Timing-tstoragefile-readActual-totalMegabytes": 104.821,
        "Timing-file-open-maxMsecs": 2.23201,
        "Timing-tstoragefile-close-minMsecs": 0.152179,
        "Timing-tstoragefile-read-minMsecs": 0.001442,
        "Timing-tstoragefile-stat-numOperations": 697,
        "Timing-root-construct-totalMsecs": 0.000643,
        "Timing-root-writev-minMsecs": 0,
        "Timing-tstoragefile-read-numSuccessfulOperations": 4737,
        "Timing-file-position-maxMsecs": 0.065989,
        "Timing-tstoragefile-close-totalMsecs": 0.992356,
        "Parameter-untracked-bool-stats": True,
        "Timing-file-writev-totalMegabytes": 0,
        "Timing-root-read-totalMegabytes": 2.65289,
        "Timing-tstoragefile-write-totalMsecs": 224.665,
        "Timing-file-flush-totalMsecs": 1847.35,
        "Timing-tstoragefile-close-numSuccessfulOperations": 4,
        "Timing-tstoragefile-seek-maxMsecs": 0.05572,
        "Timing-root-prefetch-numOperations": 30,
        "Timing-root-open-minMsecs": 2228.61,
        "Timing-root-readv-totalMsecs": 16602.5,
        "Timing-tstoragefile-writeActual-numOperations": 7307,
        "Timing-file-read-maxMsecs": 15.5135,
        "Timing-root-stagein-totalMegabytes": 0,
        "Timing-root-close-totalMegabytes": 0,
        "Timing-tstoragefile-read-maxMsecs": 258.507,
        "Timing-root-write-minMsecs": 0,
        "Timing-tstoragefile-construct-maxMsecs": 4586.16,
        "Parameter-untracked-string-readHint": "auto-detect",
        "Timing-root-stagein-maxMsecs": 4.10695,
        "Timing-tstoragefile-writeActual-maxMsecs": 12.2421,
        "Timing-file-write-totalMsecs": 220.75,
        "Timing-file-readv-totalMsecs": 0,
        "Timing-tstoragefile-readAsync-maxMsecs": 0,
        "Timing-tstoragefile-seek-numOperations": 12044,
        "Timing-tstoragefile-writeViaCache-minMsecs": 0,
        "Timing-tstoragefile-write-numSuccessfulOperations": 7307,
        "Timing-tstoragefile-seek-numSuccessfulOperations": 12044,
        "Timing-tstoragefile-readViaCache-totalMsecs": 0,
        "Timing-tstoragefile-write-maxMsecs": 12.2428,
        "Timing-file-read-totalMegabytes": 5.43017,
        "Timing-tstoragefile-stat-numSuccessfulOperations": 697,
        "Timing-root-construct-maxMsecs": 0.000261,
        "Timing-tstoragefile-construct-minMsecs": 36.3145,
        "Timing-root-writev-maxMsecs": 0,
        "Timing-tstoragefile-readActual-totalMsecs": 18677.2,
        "Timing-tstoragefile-construct-totalMsecs": 11419,
        "Timing-root-writev-totalMsecs": 0,
        "Timing-file-construct-maxMsecs": 0.000291,
        "Timing-file-prefetch-minMsecs": 0,
        "Timing-root-stagein-numOperations": 1,
        "Timing-file-close-minMsecs": 0.190476,
        "Timing-file-construct-numOperations": 2,
        "Timing-file-construct-numSuccessfulOperations": 2,
        "Timing-root-stagein-minMsecs": 4.10695,
        "Timing-root-write-totalMegabytes": 0,
        "Timing-root-read-numSuccessfulOperations": 57,
        "Timing-tstoragefile-writeActual-minMsecs": 0.005877,
        "Timing-tstoragefile-readViaCache-numOperations": 3,
        "Timing-tstoragefile-readActual-maxMsecs": 4168.78,
        "Timing-file-close-numOperations": 1,
        "Timing-file-position-numOperations": 14759,
        "Timing-tstoragefile-read-totalMsecs": 2075,
        "Timing-root-position-maxMsecs": 0.001203,
        "Timing-root-construct-minMsecs": 0.000142,
        "Timing-tstoragefile-readViaCache-numSuccessfulOperations": 0,
        "Timing-file-position-numSuccessfulOperations": 14759,
        "Timing-tstoragefile-writeViaCache-maxMsecs": 0,
        "Timing-tstoragefile-writeViaCache-totalMegabytes": 0,
        "Timing-file-write-numOperations": 7307,
        "Timing-tstoragefile-flush-totalMsecs": 1847.36,
        "Timing-root-construct-numOperations": 3,
        "Timing-tstoragefile-seek-totalMsecs": 18.3395,
        "Timing-tstoragefile-stat-totalMegabytes": 0,
        "Timing-root-prefetch-totalMsecs": 0,
        "Timing-root-prefetch-maxMsecs": 0,
        "Timing-tstoragefile-writeActual-totalMsecs": 222.926
        }
    }


def CMSSWMetrics():
    """
    Helper function to convert CMSSW_METRICS dict values to their data-types
    :return: dictionary

    Output dictionary will be in form
    {
        "SystemMemory": {
            "Active(file)": <class 'int'>.
            "KernelStack": <class 'int'>,
            ...
        },
        ...
    }
    """
    rdict = {}
    for key, mdict in CMSSW_METRICS.items():
        if key == 'XrdSiteStatistics':
            rdict[key] = XrdMetrics(mdict)
            continue
        ndict = {}
        for mkey, mval in mdict.items():
            if isinstance(mval, (int, float, str, bool)):
                ndict[mkey] = type(mval)
        rdict[key] = ndict
    return rdict


def XrdMetrics(mdict):
    """
    Convert given XrdSiteStatistics dict values to their data-types
    :param mdict: XrdSiteStatistics dict
    :return: dictionary

    The output dictionary will be in a form of
    {
        "XrdSiteStatistics": {
            "readv-numOperations": [
              {
                "site": <class 'str'>,
                "value": <class 'int'>
              },
              ...
        }
        ...
    }
    """
    rdict = {}
    for key, values in mdict.items():
        entry = values[0]
        ndict = {}
        for mkey, mval in entry.items():
            if isinstance(mval, (int, float, str, bool)):
                ndict[mkey] = type(mval)
        rdict[key] = [ndict]
    return rdict



#!/usr/bin/env python
"""
_LoadByStatus_

MySQL implementation for loading a job by scheduler status
"""


from WMCore.Database.DBFormatter import DBFormatter

class LoadByStatus(DBFormatter):
    """
    _LoadByStatus_

    Load all jobs with a certain scheduler status
    """


    sql = """SELECT rj.wmbs_id AS jobid, rj.grid_id AS gridid, rj.bulk_id AS bulkid,
               st.name AS status, rj.retry_count as retry_count, rj.id AS id, wu.cert_dn AS userdn,
               wu.group_name AS usergroup, wu.role_name AS userrole
               FROM bl_runjob rj
               INNER JOIN bl_status st ON rj.sched_status = st.id
               LEFT OUTER JOIN wmbs_users wu ON wu.id = rj.user_id
               WHERE rj.status = :complete AND st.name = :status
    """



    def execute(self, status, complete = '1', conn = None, transaction = False):
        """
        _execute_

        Create new jobs
        Expect all values in the binds field
        """

        binds = {'complete': complete, 'status': status}

        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
Poll request manager for new work
"""
__all__ = []



import time
import random
import logging
from Utils.Timers import timeFunction
from Utils.wmcoreDTools import resetWatchdogTimer, moduleName
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.WorkQueue.WorkQueueReqMgrInterface import WorkQueueReqMgrInterface

class WorkQueueManagerReqMgrPoller(BaseWorkerThread):
    """
    Polls for requests
    """
    def __init__(self, queue, config, reqMgr = None):
        """
        Initialise class members
        """
        BaseWorkerThread.__init__(self)
        self.wq = queue
        self.config = config
        if reqMgr:
            self.reqMgr = reqMgr
        else:
            self.reqMgr = WorkQueueReqMgrInterface(**self.config)
        self.previousState = {}

    def setup(self, parameters):
        """
        Called at startup - introduce random delay
             to avoid workers all starting at once
        """
        t = random.randrange(self.idleTime)
        self.logger.info('Sleeping for %d seconds before 1st loop' % t)
        time.sleep(t)

    @timeFunction
    def algorithm(self, parameters):
        """
        retrieve work from RequestManager and send updates
            """
        try:
            return self.reqMgr(self.wq)
        except Exception as ex:
            self.queue.logger.error("Error in ReqMgr loop: %s" % str(ex))

        # Reset its own watchdog timer at the end of the run cycle
        logging.info(f"Resetting {moduleName(self)} watchdog timer.")
        if resetWatchdogTimer(self):
            logging.warning(f"Failed to reset {moduleName(self)} watchdog timer. The component might be restarted soon.")



#!/usr/bin/env python
"""
_UpdateAlgoDatasetAssoc_

MySQL implementation of DBSBuffer.UpdateAlgoDatasetAssoc
"""

from WMCore.Database.DBFormatter import DBFormatter

class UpdateAlgoDatasetAssoc(DBFormatter):
    """
    _UpdateAlgoDatasetAssoc_

    Update the in_dbs column for a particular algo/dataset association.
    """
    sql = """UPDATE dbsbuffer_algo_dataset_assoc SET in_dbs = :in_dbs
               WHERE id = :id"""

    def execute(self, assocID = None, inDBS = None, conn = None,
                transaction = False):
        """
        _execute_

        Update the in_dbs column
        """
        binds = {"id": assocID, "in_dbs": inDBS}
        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)

        return



#!/usr/bin/env python
"""
Script to check the proxy validity and notify users in case it's close
to expiration.

Original author: direyes@cern.ch
"""
from __future__ import division, print_function

import getopt
import os
import re
import subprocess
import sys
import socket


HOST = socket.gethostname()
USER = os.getenv('USER')


def main(argv):
    """
    This is an utility that checks the proxy validity and sends an email in
    case the time left is lesser than --time.
    Arguments:
    --proxy     : Proxy file location (Default: X509_USER_PROXY)
    --myproxy   : Check the log term proxy in myproxy (Default: True)
    --mail      : Mail where notification should be sent (Default: USER).
    --send-mail : <True|False> Send mail notification (Default: True).
    --time      : Minimun time left [hours]. It should be an integer. (Default: 48 h).
    --verbose   : Print output messages.
    --help      : Print this menu.

    Example: python /data/amaltaro/checkProxy.py --proxy /data/certs/myproxy.pem
    --time 72 --send-mail True --mail alanmalta@gmail.com,alan.malta@cern.ch --verbose
    """

    valid = ["proxy=", "myproxy=", "mail=", "send-mail=", "time=", "verbose", "help"]
    ### // Default values
    proxy = os.getenv('X509_USER_PROXY')
    myproxy = False
    verbose = False
    mail = USER
    sendMail = True
    time = 3

    try:
        opts, _args = getopt.getopt(argv, "", valid)
    except getopt.GetoptError as ex:
        print("Options: {}\n\nException: {}".format(main.__doc__, str(ex)))
        sys.exit(1)

    ### // Handle arguments given in the command line
    for opt, arg in opts:
        if opt == "--help":
            print(main.__doc__)
            sys.exit(0)
        if opt == "--proxy":
            proxy = arg
            if proxy.startswith("~/"):
                proxy = os.getenv('HOME') + proxy[1:]
            if not os.path.exists(proxy):
                print("Proxy File does not exist")
                sys.exit(2)
        if opt == "--mail":
            mail = arg
        if opt == "--myproxy":
            myproxy = arg
        if opt == "--send-mail":
            sendMail = arg
        if opt == "--time":
            time = int(arg)
            if time < 1:
                print("Invalid time format. Check the options: {}".format(main.__doc__))
                raise sys.exit(3)
        if opt == "--verbose":
            verbose = True

    command = ["voms-proxy-info", "-file", str(proxy)]
    p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, _err = p.communicate()
    proxyInfo = [line for line in out.split(b'\n') if line]
    processTimeLeft(sendMail, verbose, proxyInfo, time, mail)

    if myproxy:
        os.environ["X509_USER_CERT"] = proxy
        os.environ["X509_USER_KEY"] = proxy
        command = ["myproxy-info", "-v", "-l", "amaltaro", "-s", "myproxy.cern.ch", "-k", "amaltaroCERN"]
        p = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = p.communicate()
        proxyInfo = [line for line in out.split(b'\n') if line]
        processTimeLeft(sendMail, verbose, proxyInfo, time, mail)


def sendMailNotification(mail, message, proxyInfo='', verbose=False):
    if verbose:
        print("Host: {}".format(HOST))
    # append proxy info to message
    for line in proxyInfo:
        message += "%s\n" % line

    # retrieve short and domain name to be used in the from-address of the email
    host_name = HOST.split('.')[0]
    domain_name = HOST.split('.')[1:]
    domain_name = '.'.join(domain_name)
    # echo %msg | mail -s 'HOST proxy status' -r FROM_ADDRESS TO_ADDRESS
    command = f" echo \"{message}\" | "
    command += f"mail -s '{HOST}: Proxy status'"
    command += f" -r {USER}-{host_name}@{domain_name}"
    command += f" {mail}"

    if verbose:
        print("Running email command: {}".format(command))
    retCode = os.system(command)
    if verbose:
        print("Email notification exit code: {}".format(retCode))


def processTimeLeft(sendMail, verbose, proxyInfo, time, mail):
    """
    Receive the whole proxy info and return its time left.
    In case no proxy information is provided, it sends an
    email warning the user.
    """
    if proxyInfo:
        if verbose:
            print('Proxy information: {}'.format(proxyInfo))
        timeLeft = []
        for line in proxyInfo:
            if line.find(b'timeleft') > -1:
                dateReg = re.compile(rb'\d{1,3}[:/]\d{2}[:/]\d{2}')
                timeLeft = dateReg.findall(line)[0]
                timeLeft = timeLeft.split(b':')[0]
                continue
    else:
        msg = "No valid proxy found in %s. " % HOST
        msg += "Please create one.\n"

        if verbose:
            print(msg)
            print("Send mail: {}".format(sendMail))

        if sendMail:
            if verbose:
                print("Sending mail notification")
            sendMailNotification(mail, msg)
        sys.exit(4)

        ### // build message
    if int(time) >= int(timeLeft):
        msg = "\nProxy file in %s is about to expire. " % HOST
        msg += "Please renew it.\n"
        msg += "Hours left: %i\n" % int(timeLeft)
        if int(timeLeft) == 0:
            msg = "Proxy file in %s HAS expired." % HOST
            msg += "Please renew it.\n"

        if verbose:
            print(msg)
            print("Send mail: {}".format(sendMail))

        ### // Sends an email
        if sendMail:
            if verbose:
                print("Sending mail notification")
            sendMailNotification(mail, msg, proxyInfo, verbose)


if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python
"""
_GetDeletedBlocksByWorkflow_

Oracle implementation of Workflow.GetDeletedBlocksByWorkflow

NOTE: This DAO is not used in the production code but is to be used only for debugging purposes
"""

from WMCore.WMBS.MySQL.Workflow.GetDeletedBlocksByWorkflow import GetDeletedBlocksByWorkflow as MySQLGetDeletedBlocksByWorkflow


class GetDeletedBlocksByWorkflow(MySQLGetDeletedBlocksByWorkflow):
    """
    Retrieves a list of all workflows and the relative deleted blocks lists
    """



"""
WorkQueueElementsSummary

"""
from __future__ import (print_function, division)

from builtins import str, bytes, object
from future.utils import viewitems

from collections import defaultdict
from math import ceil

from WMCore.WorkQueue.DataStructs.WorkQueueElement import possibleSites
from WMCore.WorkQueue.DataStructs.WorkQueueElementResult import WorkQueueElementResult


def getGlobalSiteStatusSummary(elements, status=None, dataLocality=False):
    """
    _getGlobalSiteStatusSummary_

    Given a dict with workqueue elements keyed by status, such as this format:
    {u'Canceled': [{u'Inputs': {}, u'Jobs': 18,...}, {u'Jobs': 11,...}],
     u'Running': [{'Priority': 190000,..}, ...]}

    Creates a summary of jobs and number of wq elements in each status
    distributed among the sites whitelisted. There are 2 job distribution:
     *) unique top level jobs per site and per status and (equally
        distributed among all sites)
     *) possible top level jobs per site and per status (consider all
        jobs can run in a single location)

    If status is provided, then skip any workqueue element not in the
    given status. Otherwise filter only active workqueue status.

    If dataLocality is set to True, then it considers only sites that pass
    the data location constraint.
    """
    if status and isinstance(status, (str, bytes)):
        activeStatus = [status]
    elif status and isinstance(status, (list, tuple)):
        activeStatus = status
    else:
        activeStatus = list(elements)

    uniqueJobsSummary = {}
    possibleJobsSummary = {}

    for st in activeStatus:
        uniqueJobsSummary.setdefault(st, {})
        possibleJobsSummary.setdefault(st, {})
        uniqueJobs = {}
        possibleJobs = {}
        for elem in elements.get(st, []):
            elem = elem['WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement']
            if dataLocality:
                commonSites = possibleSites(elem)
            else:
                commonSites = list(set(elem['SiteWhitelist']) - set(elem['SiteBlacklist']))

            try:
                jobsPerSite = elem['Jobs'] / len(commonSites)
            except ZeroDivisionError:
                commonSites = ['NoPossibleSite']
                jobsPerSite = elem['Jobs']

            for site in commonSites:
                uniqueJobs.setdefault(site, {'sum_jobs': 0, 'num_elem': 0})
                possibleJobs.setdefault(site, {'sum_jobs': 0, 'num_elem': 0})

                uniqueJobs[site]['sum_jobs'] += ceil(jobsPerSite)
                uniqueJobs[site]['num_elem'] += 1
                possibleJobs[site]['sum_jobs'] += ceil(elem['Jobs'])
                possibleJobs[site]['num_elem'] += 1

        uniqueJobsSummary[st].update(uniqueJobs)
        possibleJobsSummary[st].update(possibleJobs)

    return uniqueJobsSummary, possibleJobsSummary


class WorkQueueElementsSummary(object):
    """Class to hold the status of a related group of WorkQueueElements"""
    def __init__(self, elements):
        self.elements = elements

        elementsByRequest = {}
        for ele in elements:
            elementsByRequest.setdefault(ele["RequestName"], [])
            elementsByRequest[ele["RequestName"]].append(ele)

        self.wqResultsByRequest = {}
        for reqName, wqElements in viewitems(elementsByRequest):
            self.wqResultsByRequest[reqName] = WorkQueueElementResult(Elements=wqElements)

    def elementsWithHigherPriorityInSameSites(self, requestName, returnFormat="dict"):

        if requestName not in self.wqResultsByRequest:
            return None

        priority = self.wqResultsByRequest[requestName]['Priority']
        creationTime = self.wqResultsByRequest[requestName]['Elements'][0]['CreationTime']

        sites = self.getPossibleSitesByRequest(requestName)

        sortedElements = []
        for reqName in self.wqResultsByRequest:
            # skip the workflow
            if reqName == requestName:
                continue
            if self.wqResultsByRequest[reqName]['Priority'] >= priority:
                for element in self.wqResultsByRequest[reqName]['Elements']:
                    if element['CreationTime'] > creationTime:
                        continue
                    if len(sites) > 0:
                        commonSites = possibleSites(element)
                        if len(set(commonSites) & sites) > 0:
                            sortedElements.append(element)
                    else:
                        sortedElements.append(element)
        # sort elements to get them in priority first and timestamp order
        sortedElements.sort(key=lambda element: element['CreationTime'])
        sortedElements.sort(key = lambda x: x['Priority'], reverse = True)
        if returnFormat == "list":
            return sortedElements
        elif returnFormat == "dict":
            sortedByRequest = defaultdict(list)
            for ele in sortedElements:
                sortedByRequest[ele['RequestName']].append(ele)

            for request in sortedByRequest:
                sortedByRequest[request] = WorkQueueElementResult(Elements=sortedByRequest[request])
            return sortedByRequest

    def getPossibleSitesByRequest(self, requestName):

        if requestName not in self.wqResultsByRequest:
            return None
        # this will include all the possible sites on the requests
        # TODO: when different blocks are located in different site it need to handled
        sites = set()
        for ele in self.wqResultsByRequest[requestName]['Elements']:
            sites = sites | set(possibleSites(ele))
        return sites

    def getWQElementResultsByRequest(self, requestName = None):
        if requestName:
            return self.wqResultsByRequest.get(requestName, None)
        else:
            return self.wqResultsByRequest

    def printSummary(self, request, detail=False):

        wqResult = self.getWQElementResultsByRequest(request)

        if wqResult is None:
            print("No WQ element exist for the status given")
            return
        print("### summary for %s ###" % request )
        print("  Priority: %s, available elements: %s " % (wqResult["Priority"], len(wqResult['Elements'])))

        sites = self.getPossibleSitesByRequest(request)
        print("  Possible sites to run: %s" % list(sites))

        higher = self.elementsWithHigherPriorityInSameSites(request)
        total = 0
        totalJobs = 0
        for request in higher:
            wqResult = higher[request]
            availableEle = wqResult.availableItems()
            if not availableEle:
                continue
            total += len(availableEle)
            wqAvailResult = WorkQueueElementResult(Elements=availableEle)
            totalJobs += wqAvailResult['Jobs']
            maxJobEle = wqAvailResult.getMaxJobElement()
            minJobEle = wqAvailResult.getMinJobElement()
            if detail:
                print("  Higher priority elements by request:")
                print("""%s: Priority: %s, available elements: %s, total jobs: %s,
                        max jobs: %s (element_id: %s), min jobs: %s (element_id: %s)""" % (
                        request, wqAvailResult["Priority"], availableEle, wqAvailResult['Jobs'],
                        maxJobEle["Jobs"], maxJobEle.id,
                        minJobEle["Jobs"], minJobEle.id))
        print("  Higher priority elements: %s, total available jobs: %s"  % (total, totalJobs))




#!/usr/bin/env python
"""
_StageOutImpl_

Interface for Stage Out Plugins. All stage out implementations should
inherit this object and implement the methods accordingly

"""
from __future__ import print_function

from builtins import range

import logging
import os
import time

from WMCore.Storage.Execute import runCommandWithOutput
from WMCore.Storage.StageOutError import StageOutError


class StageOutImpl:
    """
    _StageOutImpl_

    Define the interface that needs to be implemented by stage out
    plugins

    Object attributes:

    - *numRetries* : Number of automated retry attempts if the command fails
                     default is 3 attempts
    - *retryPause* : Time in seconds to wait between retries.
                     default is 10 minutes
    """

    def __init__(self, stagein=False):
        self.numRetries = 3
        self.retryPause = 600
        self.stageIn = stagein
        self.debuggingTemplate =  "#!/bin/bash\n"
        self.debuggingTemplate += """
        echo
        echo
        echo "-----------------------------------------------------------"
        echo "==========================================================="
        echo
        echo "Debugging information on failing copy command"
        echo
        echo "Current date and time: $(date +"%Y-%m-%d %H:%M:%S")"
        echo "copy command which failed: {copy_command}"
        echo "Hostname:   $(hostname -f)"
        echo "OS:  $(uname -r -s)"
        echo
        echo "PYTHON environment variables:"
        env | grep ^PYTHON
        echo
        echo "LD_* environment variables:"
        env | grep ^LD_
        echo
        echo "Information for credentials in the environment"
        echo "Bearer token content: $BEARER_TOKEN"
        echo "Bearer token file: $BEARER_TOKEN_FILE"
        echo
        echo "Source PFN: {source}"
        echo "Target PFN: {destination}"
        echo
        echo "VOMS proxy info:"
        voms-proxy-info -all
        echo "==========================================================="
        echo "-----------------------------------------------------------"
        echo
        """

    def isBearerTokenFileSet(self):
        """
        Method to check if the BEARER_TOKEN_FILE environment variable is
        set and not empty. If so, whether the file exists.
        :return: True if BEARER_TOKEN_FILE envir var is set and exists, False otherwise.
        """
        if os.getenv("BEARER_TOKEN_FILE"):
            if os.path.exists(os.getenv("BEARER_TOKEN_FILE")):
                return True
        return False

    @staticmethod
    def splitPFN(pfn):
        """
        _splitPFN_

        Generic function to split the PFN in smaller pieces, such as:
        { <protocol>, <host>, <path>, <opaque> }
        """
        protocol = pfn.split(':')[0]
        host = pfn.split('/')[2]
        thisList = pfn.replace('{}://{}/'.format(protocol, host), '').split('?')
        path = thisList[0]
        opaque = ""
        # If we have any opaque info keep it
        if len(thisList) == 2:
            opaque = "?{}".format(thisList[1])

        # check for the path to actually be in the opaque information
        if opaque.startswith("?path="):
            elements = opaque.split('&')
            path = elements[0].replace('?path=', '')
            buildingOpaque = '?'
            for element in elements[1:]:
                buildingOpaque += element
                buildingOpaque += '&'
            opaque = buildingOpaque.rstrip('&')
        elif opaque.find("&path=") != -1:
            elements = opaque.split('&')
            buildingOpaque = elements[0]
            for element in elements[1:]:
                if element.startswith('path='):
                    path = element.replace('path=', '')
                else:
                    buildingOpaque += '&' + element
            opaque = buildingOpaque
        return protocol, host, path, opaque

    def executeCommand(self, command):
        """
        _execute_

        Execute the command provided, throw a StageOutError if it exits
        non zero

        """
        try:
            exitCode, output = runCommandWithOutput(command)
            msg = "Command exited with status: {}\nOutput message: {}".format(exitCode, output)
            logging.info(msg)
        except Exception as ex:
            raise StageOutError(str(ex), Command=command, ExitCode=60311) from ex

        if exitCode:
            msg = "Command exited non-zero, ExitCode: {}\nOutput: {}".format(exitCode, output)
            formatted_msg = "Exception During Stage Out:\n{}".format(msg)
            logging.error(formatted_msg)
            raise StageOutError(msg, Command=command, ExitCode=exitCode)

    def createSourceName(self, protocol, pfn):
        """
        _createSourceName_

        construct a source URL/PFN for the pfn provided based on the
        protocol that can be passed to the stage command that this
        implementation uses.

        """
        raise NotImplementedError("StageOutImpl.createSourceName")

    def createTargetName(self, protocol, pfn):
        """
        _createTargetName_

        construct a target URL/PFN for the pfn provided based on the
        protocol that can be passed to the stage command that this
        implementation uses.

        By default this is the same as createSourceName (in cases
        of stage ins the 'local' file is the target). Override this
        in your implementation of this is not the case.

        """
        return self.createSourceName(protocol, pfn)

    def createOutputDirectory(self, targetPFN):
        """
        _createOutputDirectory_

        If a seperate step is required to create a directory in the
        SE for the stage out PFN provided, do that in this command.

        If no directory is required, do not implement this method
        """

    def createStageOutCommand(self, sourcePFN, targetPFN, options=None, checksums=None, authMethod=None, forceMethod=False):
        """
        _createStageOutCommand_

        Build a shell command that will transfer the sourcePFN to the
        targetPFN using the options provided if necessary

        """
        raise NotImplementedError("StageOutImpl.createStageOutCommand")

    def createDebuggingCommand(self, sourcePFN, targetPFN, options=None, checksums=None, authMethod=None, forceMethod=False):
        """
        Build a shell command that will report in the logs the details about
        failing stageOut commands
        """
        raise NotImplementedError("StageOutImpl.createDebuggingCommand")
    
    def removeFile(self, pfnToRemove):
        """
        _removeFile_

        Construct and issue the command to remove the PFN provided as
        this impl requires.
        This will be used by the cleanup nodes in merge jobs that remove the
        intermediate files upon successful completion of the merge job

        """
        raise NotImplementedError("StageOutImpl.removeFile")

    def createRemoveFileCommand(self, pfn):
        """
        return the command to delete a file after a failed copy
        """
        if pfn.startswith("/"):
            return "/bin/rm -f {}".format(pfn)
        elif os.path.isfile(pfn):
            return "/bin/rm -f {}".format(os.path.abspath(pfn))
        else:
            return ""

    def __call__(self, protocol, inputPFN, targetPFN, options=None, checksums=None):
        """
        _Operator()_

        This operator does the actual stage out by invoking the overridden
        plugin methods of the derived object.
        """

        # //
        # // Generate the source PFN from the plain PFN if needed
        # //
        sourcePFN = self.createSourceName(protocol, inputPFN)

        # destination may also need PFN changed
        # i.e. if we are staging in a file from an SE
        targetPFN = self.createTargetName(protocol, targetPFN)
        # //
        # // Create the output directory if implemented
        # //
        for retryCount in range(self.numRetries + 1):
            try:
                logging.info("Creating output directory...")
                self.createOutputDirectory(targetPFN)
                break
            except StageOutError as ex:
                msg = "Attempt {} to create a directory for stageout failed.\n".format(retryCount)
                msg += "Automatically retrying stage out in {} secs\n ".format(self.retryPause)
                msg += "Error details:\n{}\n".format(str(ex))
                logging.error(msg)
                if retryCount == self.numRetries:
                    # //
                    # // last retry, propagate exception
                    # //
                    logging.error("Maximum retries exhausted when trying to create the output directory")
                    raise ex
                time.sleep(self.retryPause)

        # //
        # // Create the command to be used.
        # //
        # // This will actually only enforce the definition of the bearer token variables, still x509 auth activated as a backup
        command = self.createStageOutCommand(sourcePFN, targetPFN, options, checksums, authMethod="TOKEN")
        # //
        # // Run the command
        # //

        stageOutEx = None  # variable to store the possible StageOutError
        for retryCount in range(self.numRetries + 1):
            try:
                logging.info("Running the stage out with the available auth method (attempt %d)...", retryCount + 1)
                logging.info("Command to run: %s", command)
                self.executeCommand(command)
                logging.info("\nStage-out succeeded with the current environment.")
                break

            except StageOutError as ex:
                msg = "Attempt {} to stage out failed with default setup.\n".format(retryCount)
                msg += "Error details:\n{}\n".format(str(ex))
                logging.error(msg)
                logging.info("Retrying with authentication-safe logic...")

                # Authentication-safe fallback logic
                if os.getenv("X509_USER_PROXY"):
                    logging.info("Retrying with X509_USER_PROXY with BEARER_TOKEN unset...")
                    command = self.createStageOutCommand(sourcePFN, targetPFN, options, checksums, authMethod="X509", forceMethod=True)
                    try:
                        logging.info("Command to run: %s", command)
                        self.executeCommand(command)
                        logging.info("Stage-out succeeded with X509 with BEARER_TOKEN unset.")
                        return
                    except StageOutError as fallbackEx:
                        logging.warning("Fallback with X509_USER_PROXY failed:\n%s", str(fallbackEx))

                if os.getenv("BEARER_TOKEN") or os.getenv("BEARER_TOKEN_FILE"):
                    logging.info("Retrying with BEARER_TOKEN with X509_USER_PROXY unset...")
                    command = self.createStageOutCommand(sourcePFN, targetPFN, options, checksums, authMethod="TOKEN", forceMethod=True)
                    try:
                        logging.info("Command to run: %s", command)
                        self.executeCommand(command)
                        logging.info("Stage-out succeeded with TOKEN with X509_USER_PROXY unset.")
                        return
                    except StageOutError as fallbackEx:
                        logging.warning("Fallback with BEARER_TOKEN failed:\n%s", str(fallbackEx))

                if retryCount == self.numRetries:
                    # Last retry, propagate the information outside of the for loop
                    stageOutEx = ex
                msg += "Automatically retrying in {} secs\n ".format(self.retryPause)
                time.sleep(self.retryPause)

        # This block will now always be executed after retries are exhausted
        if stageOutEx is not None:
            logging.error("Maximum number of retries exhausted. Further details on the failed command reported below.")
            command = self.createDebuggingCommand(sourcePFN, targetPFN, options, checksums, authMethod="TOKEN")
            self.executeCommand(command)
            raise stageOutEx from None



from __future__ import (division, print_function)
from time import time
from WMCore.REST.CherryPyPeriodicTask import CherryPyPeriodicTask
from WMCore.WorkQueue.WorkQueue import globalQueue

class LocationUpdateTask(CherryPyPeriodicTask):

    def __init__(self, rest, config):

        super(LocationUpdateTask, self).__init__(config)
        self.globalQ = globalQueue(logger=self.logger, **config.queueParams)

    def setConcurrentTasks(self, config):
        """
        sets the list of function reference for concurrent tasks
        """
        self.concurrentTasks = [{'func': self.updateDataLocation, 'duration': config.locationUpdateDuration}]

    def updateDataLocation(self, config):
        """
        gather active data statistics
        """
        tStart = time()
        res = self.globalQ.updateLocationInfo()
        self.logger.info("%s executed in %.3f secs and updated %d non-unique elements",
                         self.__class__.__name__, time() - tStart, res)

        return



#!/usr/bin/env python
"""
updateLocations poller
"""
__all__ = []


import time
import random
import logging

from Utils.Timers import timeFunction
from Utils.wmcoreDTools import resetWatchdogTimer, moduleName
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread


class WorkQueueManagerLocationPoller(BaseWorkerThread):
    """
    Polls for location updates
    """
    def __init__(self, queue, config):
        """
        Initialise class members
        """
        BaseWorkerThread.__init__(self)
        self.queue = queue
        self.config = config

    def setup(self, parameters):
        """
        Called at startup - introduce random delay
             to avoid workers all starting at once
        """
        t = random.randrange(self.idleTime)
        self.logger.info('Sleeping for %d seconds before 1st loop' % t)
        time.sleep(t)

    @timeFunction
    def algorithm(self, parameters):
        """
        Update locations
            """

        self.queue.logger.info("Updating data locations")
        try:
            self.queue.updateLocationInfo()
        except Exception as ex:
            self.queue.logger.error("Error updating locations: %s" % str(ex))

        # Reset its own watchdog timer at the end of the run cycle
        logging.info(f"Resetting {moduleName(self)} watchdog timer.")
        if resetWatchdogTimer(self):
            logging.warning(f"Failed to reset {moduleName(self)} watchdog timer. The component might be restarted soon.")
